# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Helper functions to prepare ASTRA algorithms.

This module contains utility functions to convert data structures from the
ODL geometry representation to ASTRA's data structures, including:

* volume geometries
* projection geometries
* create vectors from geometries
* data arrays
* projectors
* algorithm configuration dictionaries

`ASTRA documentation <http://www.astra-toolbox.com/>`_.

`ASTRA on GitHub <https://github.com/astra-toolbox/>`_.
"""

from __future__ import absolute_import, division, print_function

import warnings

import numpy as np

from odl.discr import DiscretizedSpace, DiscretizedSpaceElement
from odl.tomo.geometry import (
    DivergentBeamGeometry, Flat1dDetector, Flat2dDetector, Geometry,
    ParallelBeamGeometry)
from odl.tomo.util.utility import euler_matrix

try:
    import astra
except ImportError:
    ASTRA_AVAILABLE = False
    ASTRA_VERSION = ''
else:
    ASTRA_AVAILABLE = True

# Make sure that ASTRA >= 1.7 is used
if ASTRA_AVAILABLE:
    try:
        # Available from 1.8 on
        ASTRA_VERSION = astra.__version__
    except AttributeError:
        # Below version 1.8
        _maj = astra.astra.version() // 100
        _min = astra.astra.version() % 100
        ASTRA_VERSION = '.'.join([str(_maj), str(_min)])
        if (_maj, _min) < (1, 7):
            warnings.warn(
                'your version {}.{} of ASTRA is unsupported, please upgrade '
                'to 1.7 or higher'.format(_maj, _min), RuntimeWarning)

__all__ = (
    'ASTRA_AVAILABLE',
    'ASTRA_VERSION',
    'astra_supports',
    'astra_versions_supporting',
    'astra_volume_geometry',
    'astra_conebeam_3d_geom_to_vec',
    'astra_conebeam_2d_geom_to_vec',
    'astra_parallel_3d_geom_to_vec',
    'astra_projection_geometry',
    'astra_data',
    'astra_projector',
    'astra_algorithm',
)


# ASTRA_FEATURES contains a set of features along with version specifiers
# to track ASTRA support for those features. The version specifiers must
# be valid setuptools package requirements (without package name), e.g.,
# as exact version '==1.7', as inequality '>=1.7' or as several requirements
# that need to be satisfied simultaneously, e.g., '>=1.8,<=2.0'.
# To give multiple requirements that should be OR-ed together, use a
# sequence instead of a single string in the dictionary, e.g.,
# ['==1.7', '>=1.8,<=2.0'].

ASTRA_FEATURES = {
    # Cell sizes not equal in both axes in 2d, currently crashes.
    'anisotropic_voxels_2d': None,

    # Cell sizes not equal all 3 axes in 3d, see
    # https://github.com/astra-toolbox/astra-toolbox/pull/41
    'anisotropic_voxels_3d': '>=1.8',

    # ASTRA geometry defined by vectors supported in the current
    # development version, will be in the next release after 1.8.3. See
    # https://github.com/astra-toolbox/astra-toolbox/issues/54
    'par2d_vec_geometry': '>1.8.3',

    # Density weighting for cone 2d (fan beam), not supported yet,
    # see the discussion in
    # https://github.com/astra-toolbox/astra-toolbox/issues/71
    'cone2d_density_weighting': None,

    # Approximate version of ray-density weighting in cone beam
    # backprojection for constant source-detector distance, see
    # https://github.com/astra-toolbox/astra-toolbox/pull/84
    'cone3d_approx_density_weighting': '>=1.8,<1.9.9.dev',

    # General case not supported yet, see the discussion in
    # https://github.com/astra-toolbox/astra-toolbox/issues/71
    'cone3d_density_weighting': None,

    # Fix for division by zero with detector midpoint normal perpendicular
    # to geometry axis in parallel 3d, see
    # https://github.com/astra-toolbox/astra-toolbox/issues/18
    'par3d_det_mid_pt_perp_to_axis': '>=1.7.2',

    # Linking instead of copying of GPU memory, see
    # https://github.com/astra-toolbox/astra-toolbox/pull/93
    'gpulink': '>=1.8.3',

    # Distance-driven projector for parallel 2d geometry, will be in the
    # next release after 1.8.3, see
    # https://github.com/astra-toolbox/astra-toolbox/pull/183
    'par2d_distance_driven_proj': '>1.8.3',
}


def astra_supports(feature):
    """Return bool indicating whether current ASTRA supports ``feature``.

    Parameters
    ----------
    feature : str
        Name of a potential feature of ASTRA. See ``ASTRA_FEATURES`` for
        possible values.

    Returns
    -------
    supports : bool
        ``True`` if the currently imported version of ASTRA supports the
        feature in question, ``False`` otherwise.
    """
    from odl.util.utility import pkg_supports
    return pkg_supports(feature, ASTRA_VERSION, ASTRA_FEATURES)


def astra_versions_supporting(feature):
    """Return version spec for support of the given feature.

    Parameters
    ----------
    feature : str
        Name of a potential feature of ASTRA. See ``ASTRA_FEATURES`` for
        possible values.

    Returns
    -------
    version_spec : str
        Specifier for versions of ASTRA that support ``feature``. See
        `odl.util.utility.pkg_supports` for details.
    """
    try:
        return ASTRA_FEATURES[str(feature)]
    except KeyError:
        raise ValueError('unknown feature {!r}'.format(feature))


def astra_volume_geometry(vol_space):
    """Create an ASTRA volume geometry from the discretized domain.

    From the ASTRA documentation:

    In all 3D geometries, the coordinate system is defined around the
    reconstruction volume. The center of the reconstruction volume is the
    origin, and the sides of the voxels in the volume have length 1.

    All dimensions in the projection geometries are relative to this unit
    length.


    Parameters
    ----------
    vol_space : `DiscretizedSpace`
        Discretized space where the reconstruction (volume) lives.
        It must be 2- or 3-dimensional and uniformly discretized.

    Returns
    -------
    astra_geom : dict

    Raises
    ------
    NotImplementedError
        If the cell sizes are not the same in each dimension.
    """
    if not isinstance(vol_space, DiscretizedSpace):
        raise TypeError('`vol_space` {!r} is not a DiscretizedSpace instance'
                        ''.format(vol_space))

    if not vol_space.is_uniform:
        raise ValueError('`vol_space` {} is not uniformly discretized')

    vol_shp = vol_space.partition.shape
    vol_min = vol_space.partition.min_pt
    vol_max = vol_space.partition.max_pt

    if vol_space.ndim == 2:
        # ASTRA does in principle support custom minimum and maximum
        # values for the volume extent also in earlier versions, but running
        # the algorithm fails if voxels are non-isotropic.
        if (
            not vol_space.partition.has_isotropic_cells
            and not astra_supports('anisotropic_voxels_2d')
        ):
            req_ver = astra_versions_supporting('anisotropic_voxels_2d')
            raise NotImplementedError(
                'support for non-isotropic pixels in 2d volumes requires '
                'ASTRA {}'.format(req_ver)
            )
        # Given a 2D array of shape (x, y), a volume geometry is created as:
        #    astra.create_vol_geom(x, y, y_min, y_max, x_min, x_max)
        # yielding a dictionary:
        #   {'GridRowCount': x,
        #    'GridColCount': y,
        #    'WindowMinX': y_min,
        #    'WindowMaxX': y_max,
        #    'WindowMinY': x_min,
        #    'WindowMaxY': x_max}
        #
        # NOTE: this setting is flipped with respect to x and y. We do this
        # as part of a global rotation of the geometry by -90 degrees, which
        # avoids rotating the data.
        # NOTE: We need to flip the sign of the (ODL) x component since
        # ASTRA seems to move it in the other direction. Not quite clear
        # why.
        vol_geom = astra.create_vol_geom(vol_shp[0], vol_shp[1],
                                         vol_min[1], vol_max[1],
                                         -vol_max[0], -vol_min[0])
    elif vol_space.ndim == 3:
        # Not supported in all versions of ASTRA
        if (
            not vol_space.partition.has_isotropic_cells
            and not astra_supports('anisotropic_voxels_3d')
        ):
            req_ver = astra_versions_supporting('anisotropic_voxels_3d')
            raise NotImplementedError(
                'support for non-isotropic pixels in 3d volumes requires '
                'ASTRA {}'.format(req_ver)
            )
        # Given a 3D array of shape (x, y, z), a volume geometry is created as:
        #    astra.create_vol_geom(y, z, x, z_min, z_max, y_min, y_max,
        #                          x_min, x_max),
        # yielding a dictionary:
        #   {'GridColCount': z,
        #    'GridRowCount': y,
        #    'GridSliceCount': x,
        #    'WindowMinX': z_max,
        #    'WindowMaxX': z_max,
        #    'WindowMinY': y_min,
        #    'WindowMaxY': y_min,
        #    'WindowMinZ': x_min,
        #    'WindowMaxZ': x_min}
        vol_geom = astra.create_vol_geom(vol_shp[1], vol_shp[2], vol_shp[0],
                                         vol_min[2], vol_max[2],
                                         vol_min[1], vol_max[1],
                                         vol_min[0], vol_max[0])
    else:
        raise ValueError('{}-dimensional volume geometries not supported '
                         'by ASTRA'.format(vol_space.ndim))
    return vol_geom


def astra_conebeam_3d_geom_to_vec(geometry):
    """Create vectors for ASTRA projection geometries from ODL geometry.

    The 3D vectors are used to create an ASTRA projection geometry for
    cone beam geometries, see ``'cone_vec'`` in the
    `ASTRA projection geometry documentation`_.

    Each row of the returned vectors corresponds to a single projection
    and consists of ::

        (srcX, srcY, srcZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ)

    with

        - ``src``: the ray source position
        - ``d``  : the center of the detector
        - ``u``  : the vector from detector pixel ``(0,0)`` to ``(0,1)``
        - ``v``  : the vector from detector pixel ``(0,0)`` to ``(1,0)``

    Parameters
    ----------
    geometry : `Geometry`
        ODL projection geometry from which to create the ASTRA geometry.

    Returns
    -------
    vectors : `numpy.ndarray`
        Array of shape ``(num_angles, 12)`` containing the vectors.

    References
    ----------
    .. _ASTRA projection geometry documentation:
       http://www.astra-toolbox.com/docs/geom3d.html#projection-geometries
    """
    angles = geometry.angles
    vectors = np.zeros((angles.size, 12))

    # Source position
    vectors[:, 0:3] = geometry.src_position(angles)

    # Center of detector in 3D space
    mid_pt = geometry.det_params.mid_pt
    vectors[:, 3:6] = geometry.det_point_position(angles, mid_pt)

    # Vectors from detector pixel (0, 0) to (1, 0) and (0, 0) to (0, 1)
    # `det_axes` gives shape (N, 2, 3), swap to get (2, N, 3)
    det_axes = np.moveaxis(geometry.det_axes(angles), -2, 0)
    px_sizes = geometry.det_partition.cell_sides
    # Swap detector axes to have better memory layout in  projection data.
    # ASTRA produces `(v, theta, u)` layout, and to map to ODL layout
    # `(theta, u, v)` a complete roll must be performed, which is the
    # worst case (compeltely discontiguous).
    # Instead we swap `u` and `v`, resulting in the effective ASTRA result
    # `(u, theta, v)`. Here we only need to swap axes 0 and 1, which
    # keeps at least contiguous blocks in `v`.
    vectors[:, 9:12] = det_axes[0] * px_sizes[0]
    vectors[:, 6:9] = det_axes[1] * px_sizes[1]

    # ASTRA has (z, y, x) axis convention, in contrast to (x, y, z) in ODL,
    # so we need to adapt to this by changing the order.
    newind = []
    for i in range(4):
        newind += [2 + 3 * i, 1 + 3 * i, 0 + 3 * i]
    vectors = vectors[:, newind]

    return vectors


def astra_conebeam_2d_geom_to_vec(geometry):
    """Create vectors for ASTRA projection geometries from ODL geometry.

    The 2D vectors are used to create an ASTRA projection geometry for
    fan beam geometries, see ``'fanflat_vec'`` in the
    `ASTRA projection geometry documentation`_.

    Each row of the returned vectors corresponds to a single projection
    and consists of ::

        (srcX, srcY, dX, dY, uX, uY)

    with

        - ``src``: the ray source position
        - ``d``  : the center of the detector
        - ``u``  : the vector from detector pixel 0 to 1

    Parameters
    ----------
    geometry : `Geometry`
        ODL projection geometry from which to create the ASTRA geometry.

    Returns
    -------
    vectors : `numpy.ndarray`
        Array of shape ``(num_angles, 6)`` containing the vectors.

    References
    ----------
    .. _ASTRA projection geometry documentation:
       http://www.astra-toolbox.com/docs/geom2d.html#projection-geometries
    """
    # Instead of rotating the data by 90 degrees counter-clockwise,
    # we subtract pi/2 from the geometry angles, thereby rotating the
    # geometry by 90 degrees clockwise
    rot_minus_90 = euler_matrix(-np.pi / 2)
    angles = geometry.angles
    vectors = np.zeros((angles.size, 6))

    # Source position
    src_pos = geometry.src_position(angles)
    vectors[:, 0:2] = rot_minus_90.dot(src_pos.T).T  # dot along 2nd axis

    # Center of detector
    mid_pt = geometry.det_params.mid_pt
    # Need to cast `mid_pt` to float since otherwise the empty axis is
    # not removed
    centers = geometry.det_point_position(angles, float(mid_pt))
    vectors[:, 2:4] = rot_minus_90.dot(centers.T).T

    # Vector from detector pixel 0 to 1
    det_axis = rot_minus_90.dot(geometry.det_axis(angles).T).T
    px_size = geometry.det_partition.cell_sides[0]
    vectors[:, 4:6] = det_axis * px_size

    return vectors


def astra_parallel_3d_geom_to_vec(geometry):
    """Create vectors for ASTRA projection geometries from ODL geometry.

    The 3D vectors are used to create an ASTRA projection geometry for
    parallel beam geometries, see ``'parallel3d_vec'`` in the
    `ASTRA projection geometry documentation`_.

    Each row of the returned vectors corresponds to a single projection
    and consists of ::

        (rayX, rayY, rayZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ)

    with

        - ``ray``: the ray direction
        - ``d``  : the center of the detector
        - ``u``  : the vector from detector pixel ``(0,0)`` to ``(0,1)``
        - ``v``  : the vector from detector pixel ``(0,0)`` to ``(1,0)``

    Parameters
    ----------
    geometry : `Geometry`
        ODL projection geometry from which to create the ASTRA geometry.

    Returns
    -------
    vectors : `numpy.ndarray`
        Array of shape ``(num_angles, 12)`` containing the vectors.

    References
    ----------
    .. _ASTRA projection geometry documentation:
       http://www.astra-toolbox.com/docs/geom3d.html#projection-geometries
    """
    angles = geometry.angles
    mid_pt = geometry.det_params.mid_pt

    vectors = np.zeros((angles.shape[-1], 12))

    # Ray direction = -(detector-to-source normal vector)
    vectors[:, 0:3] = -geometry.det_to_src(angles, mid_pt)

    # Center of the detector in 3D space
    vectors[:, 3:6] = geometry.det_point_position(angles, mid_pt)

    # Vectors from detector pixel (0, 0) to (1, 0) and (0, 0) to (0, 1)
    # `det_axes` gives shape (N, 2, 3), swap to get (2, N, 3)
    det_axes = np.moveaxis(geometry.det_axes(angles), -2, 0)
    px_sizes = geometry.det_partition.cell_sides
    # Swap detector axes to have better memory layout in  projection data.
    # ASTRA produces `(v, theta, u)` layout, and to map to ODL layout
    # `(theta, u, v)` a complete roll must be performed, which is the
    # worst case (compeltely discontiguous).
    # Instead we swap `u` and `v`, resulting in the effective ASTRA result
    # `(u, theta, v)`. Here we only need to swap axes 0 and 1, which
    # keeps at least contiguous blocks in `v`.
    vectors[:, 9:12] = det_axes[0] * px_sizes[0]
    vectors[:, 6:9] = det_axes[1] * px_sizes[1]

    # ASTRA has (z, y, x) axis convention, in contrast to (x, y, z) in ODL,
    # so we need to adapt to this by changing the order.
    new_ind = []
    for i in range(4):
        new_ind += [2 + 3 * i, 1 + 3 * i, 0 + 3 * i]
    vectors = vectors[:, new_ind]
    return vectors


def astra_projection_geometry(geometry):
    """Create an ASTRA projection geometry from an ODL geometry object.

    As of ASTRA version 1.7, the length values are not required any more to be
    rescaled for 3D geometries and non-unit (but isotropic) voxel sizes.

    Parameters
    ----------
    geometry : `Geometry`
        ODL projection geometry from which to create the ASTRA geometry.

    Returns
    -------
    proj_geom : dict
        Dictionary defining the ASTRA projection geometry.
    """
    if not isinstance(geometry, Geometry):
        raise TypeError('`geometry` {!r} is not a `Geometry` instance'
                        ''.format(geometry))

    if 'astra' in geometry.implementation_cache:
        # Shortcut, reuse already computed value.
        return geometry.implementation_cache['astra']

    if not geometry.det_partition.is_uniform:
        raise ValueError('non-uniform detector sampling is not supported')

    if (isinstance(geometry, ParallelBeamGeometry) and
            isinstance(geometry.detector, (Flat1dDetector, Flat2dDetector)) and
            geometry.ndim == 2):
        # TODO: change to parallel_vec when available
        det_width = geometry.det_partition.cell_sides[0]
        det_count = geometry.detector.size
        # Instead of rotating the data by 90 degrees counter-clockwise,
        # we subtract pi/2 from the geometry angles, thereby rotating the
        # geometry by 90 degrees clockwise
        angles = geometry.angles - np.pi / 2
        proj_geom = astra.create_proj_geom('parallel', det_width, det_count,
                                           angles)

    elif (isinstance(geometry, DivergentBeamGeometry) and
          isinstance(geometry.detector, (Flat1dDetector, Flat2dDetector)) and
          geometry.ndim == 2):
        det_count = geometry.detector.size
        vec = astra_conebeam_2d_geom_to_vec(geometry)
        proj_geom = astra.create_proj_geom('fanflat_vec', det_count, vec)

    elif (isinstance(geometry, ParallelBeamGeometry) and
          isinstance(geometry.detector, (Flat1dDetector, Flat2dDetector)) and
          geometry.ndim == 3):
        # Swap detector axes (see astra_*_3d_to_vec)
        det_row_count = geometry.det_partition.shape[0]
        det_col_count = geometry.det_partition.shape[1]
        vec = astra_parallel_3d_geom_to_vec(geometry)
        proj_geom = astra.create_proj_geom('parallel3d_vec', det_row_count,
                                           det_col_count, vec)

    elif (isinstance(geometry, DivergentBeamGeometry) and
          isinstance(geometry.detector, (Flat1dDetector, Flat2dDetector)) and
          geometry.ndim == 3):
        # Swap detector axes (see astra_*_3d_to_vec)
        det_row_count = geometry.det_partition.shape[0]
        det_col_count = geometry.det_partition.shape[1]
        vec = astra_conebeam_3d_geom_to_vec(geometry)
        proj_geom = astra.create_proj_geom('cone_vec', det_row_count,
                                           det_col_count, vec)
    else:
        raise NotImplementedError('unknown ASTRA geometry type {!r}'
                                  ''.format(geometry))

    if 'astra' not in geometry.implementation_cache:
        # Save computed value for later
        geometry.implementation_cache['astra'] = proj_geom

    return proj_geom


def astra_data(astra_geom, datatype, data=None, ndim=2, allow_copy=False):
    """Create an ASTRA data object.

    Parameters
    ----------
    astra_geom : dict
        ASTRA geometry object for the data creator, must correspond to the
        given ``datatype``.
    datatype : {'volume', 'projection'}
        Type of the data container.
    data : `DiscretizedSpaceElement` or `numpy.ndarray`, optional
        Data for the initialization of the data object. If ``None``,
        an ASTRA data object filled with zeros is created.
    ndim : {2, 3}, optional
        Dimension of the data. If ``data`` is provided, this parameter
        has no effect.
    allow_copy : `bool`, optional
        If ``True``, allow copying of ``data``. This means that anything
        written by ASTRA to the returned object will not be written to
        ``data``.

    Returns
    -------
    id : int
        Handle for the new ASTRA internal data object.
    """
    if data is not None:
        if isinstance(data, (DiscretizedSpaceElement, np.ndarray)):
            ndim = data.ndim
        else:
            raise TypeError('`data` {!r} is neither DiscretizedSpaceElement '
                            'instance nor a `numpy.ndarray`'.format(data))
    else:
        ndim = int(ndim)

    if datatype == 'volume':
        astra_dtype_str = '-vol'
    elif datatype == 'projection':
        astra_dtype_str = '-sino'
    else:
        raise ValueError('`datatype` {!r} not understood'.format(datatype))

    # Get the functions from the correct module
    if ndim == 2:
        link = astra.data2d.link
        create = astra.data2d.create
    elif ndim == 3:
        link = astra.data3d.link
        create = astra.data3d.create
    else:
        raise ValueError('{}-dimensional data not supported'
                         ''.format(ndim))

    # ASTRA checks if data is c-contiguous and aligned
    if data is not None:
        if allow_copy:
            data_array = np.asarray(data, dtype='float32', order='C')
            return link(astra_dtype_str, astra_geom, data_array)
        else:
            if isinstance(data, np.ndarray):
                return link(astra_dtype_str, astra_geom, data)
            elif data.tensor.impl == 'numpy':
                return link(astra_dtype_str, astra_geom, data.asarray())
            else:
                # Something else than NumPy data representation
                raise NotImplementedError('ASTRA supports data wrapping only '
                                          'for `numpy.ndarray` instances, got '
                                          '{!r}'.format(data))
    else:
        return create(astra_dtype_str, astra_geom)


def astra_projector(astra_proj_type, astra_vol_geom, astra_proj_geom, ndim):
    """Create an ASTRA projector configuration dictionary.

    Parameters
    ----------
    astra_proj_type : str
        ASTRA projector type. Available selections depend on the type of
        geometry. See `the ASTRA documentation
        <http://www.astra-toolbox.com/docs/proj2d.html>`_ for details.
    astra_vol_geom : dict
        ASTRA volume geometry.
    astra_proj_geom : dict
        ASTRA projection geometry.
    ndim : {2, 3}
        Number of dimensions of the projector.

    Returns
    -------
    proj_id : int
        Handle for the created ASTRA internal projector object.
    """
    if 'type' not in astra_proj_geom:
        raise ValueError('invalid projection geometry dict {}'
                         ''.format(astra_proj_geom))

    ndim = int(ndim)

    astra_geom = astra_proj_geom['type']
    if (
        astra_geom == 'parallel_vec'
        and not astra_supports('par2d_vec_geometry')
    ):
        req_ver = astra_versions_supporting('par2d_vec_geometry')
        raise ValueError(
            "'parallel_vec' geometry requires ASTRA {}".format(req_ver)
        )

    # Check if projector types are valid. We should not have to do this,
    # but the errors from ASTRA are rather unspecific, so we check ourselves
    # to know what's wrong.
    astra_proj_type = str(astra_proj_type).lower()

    if (
        astra_proj_type == 'distance_driven'
        and not astra_supports('par2d_distance_driven_proj')
    ):
        req_ver = astra_versions_supporting('par2d_distance_driven_proj')
        raise ValueError(
            "'distance_driven' projector requires ASTRA {}".format(req_ver)
        )

    if astra_geom in {'parallel', 'parallel_vec'}:
        valid_proj_types = ['line', 'linear', 'strip', 'cuda']
        if astra_supports('par2d_distance_driven_proj'):
            valid_proj_types.append('distance_driven')
    elif astra_geom in {'fanflat', 'fanflat_vec'}:
        valid_proj_types = ['line_fanflat', 'strip_fanflat', 'cuda']
    elif astra_geom in {'parallel3d', 'parallel3d_vec'}:
        valid_proj_types = ['linear3d', 'cuda3d']
    elif astra_geom in {'cone', 'cone_vec'}:
        valid_proj_types = ['linearcone', 'cuda3d']
    else:
        raise ValueError('invalid geometry type {!r}'.format(astra_geom))

    if astra_proj_type not in valid_proj_types:
        raise ValueError(
            'projector type {!r} not in the set {} of valid types for '
            'geometry type {!r}'
            ''.format(astra_proj_type, valid_proj_types, astra_geom)
        )

    # Create config dict
    proj_cfg = {}
    proj_cfg['type'] = astra_proj_type
    proj_cfg['VolumeGeometry'] = astra_vol_geom
    proj_cfg['ProjectionGeometry'] = astra_proj_geom
    proj_cfg['options'] = {}

    # Add the approximate 1/r^2 weighting exposed in intermediate versions of
    # ASTRA
    if (
        astra_geom in ('cone', 'cone_vec')
        and astra_supports('cone3d_approx_density_weighting')
    ):
        proj_cfg['options']['DensityWeighting'] = True

    if ndim == 2:
        return astra.projector.create(proj_cfg)
    else:
        return astra.projector3d.create(proj_cfg)


def astra_algorithm(direction, ndim, vol_id, sino_id, proj_id, impl):
    """Create an ASTRA algorithm object to run the projector.

    Parameters
    ----------
    direction : {'forward', 'backward'}
        For ``'forward'``, apply the forward projection, for ``'backward'``
        the backprojection.
    ndim : {2, 3}
        Number of dimensions of the projector.
    vol_id : int
        Handle for the ASTRA volume data object.
    sino_id : int
        Handle for the ASTRA projection data object.
    proj_id : int
        Handle for the ASTRA projector object.
    impl : {'cpu', 'cuda'}
        Implementation of the projector.

    Returns
    -------
    id : int
        Handle for the created ASTRA internal algorithm object.
    """
    if direction not in ('forward', 'backward'):
        raise ValueError("`direction` '{}' not understood".format(direction))
    if ndim not in (2, 3):
        raise ValueError('{}-dimensional projectors not supported'
                         ''.format(ndim))
    if impl not in ('cpu', 'cuda'):
        raise ValueError("`impl` type '{}' not understood"
                         ''.format(impl))
    if ndim == 3 and impl == 'cpu':
        raise NotImplementedError(
            '3d algorithms for CPU not supported by ASTRA')
    if proj_id is None and impl == 'cpu':
        raise ValueError("'cpu' implementation requires projector ID")

    algo_map = {'forward': {2: {'cpu': 'FP', 'cuda': 'FP_CUDA'},
                            3: {'cpu': None, 'cuda': 'FP3D_CUDA'}},
                'backward': {2: {'cpu': 'BP', 'cuda': 'BP_CUDA'},
                             3: {'cpu': None, 'cuda': 'BP3D_CUDA'}}}

    algo_cfg = {'type': algo_map[direction][ndim][impl],
                'ProjectorId': proj_id,
                'ProjectionDataId': sino_id}
    if direction == 'forward':
        algo_cfg['VolumeDataId'] = vol_id
    else:
        algo_cfg['ReconstructionDataId'] = vol_id

    # Create ASTRA algorithm object
    return astra.algorithm.create(algo_cfg)


if __name__ == '__main__':
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Source and detector shifts for divergent beam geometries."""

from __future__ import print_function, division, absolute_import
import numpy as np
from odl.discr.discr_utils import nearest_interpolator

__all__ = ('flying_focal_spot',)


def flying_focal_spot(angle, apart, shifts):
    """Flying focal spot shifts for divergent beam geometries.
    Shifts are defined only for grid points of angular partition.
    For all other angles nearest neighbor interpolation is used.

    Parameters
    ----------
    angle : float or `array-like`
        Angle(s) in radians describing the counter-clockwise
        rotation of source and detector.
    apart : 1-dim. `RectPartition`
        Partition of the angle interval.
    shifts : sequence of `array-like`
        Each vectors in a sequence represent a subsequent shift
        relative to the default source position. Vector elements
        represent shifts along the following directions:
        det_to_src, tangent to the rotation
        (projected on a plane perpendicular to rotation axis), rotation axis.
    """
    assert apart.ndim == 1

    angle = np.array(angle, dtype=float, copy=False, ndmin=1)
    assert angle.ndim == 1

    shifts = np.array(shifts, dtype=float, ndmin=2)
    if shifts.shape[1] not in [2, 3]:
        raise ValueError('Flying focal spot shifts must have '
                         'shape (2,) or (3,), got {}'.format(shifts))

    interpolator = nearest_interpolator(np.arange(apart.size),
                                        apart.coord_vectors)
    ind = interpolator(angle)

    k = len(shifts)
    result = [shifts[int(i) % k] for i in ind]
    return np.array(result, dtype=float, ndmin=2)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Universal functions (ufuncs) for ODL-wrapped arrays.

These functions are internal and should only be used as methods on
`Tensor`-like classes.

See `numpy.ufuncs
<http://docs.scipy.org/doc/numpy/reference/ufuncs.html>`_
for more information.

Notes
-----
The default implementation of these methods uses the ``__array_ufunc__``
dispatch machinery `introduced in Numpy 1.13
<https://github.com/charris/numpy/blob/master/doc/source/reference/\
arrays.classes.rst#special-attributes-and-methods>`_.
"""

from __future__ import print_function, division, absolute_import
from builtins import object
import numpy as np
import re


__all__ = ('TensorSpaceUfuncs', 'ProductSpaceUfuncs')


# Some are ignored since they don't cooperate with dtypes, needs fix
RAW_UFUNCS = ['absolute', 'add', 'arccos', 'arccosh', 'arcsin', 'arcsinh',
              'arctan', 'arctan2', 'arctanh', 'bitwise_and', 'bitwise_or',
              'bitwise_xor', 'ceil', 'conj', 'copysign', 'cos', 'cosh',
              'deg2rad', 'divide', 'equal', 'exp', 'exp2', 'expm1', 'floor',
              'floor_divide', 'fmax', 'fmin', 'fmod', 'greater',
              'greater_equal', 'hypot', 'invert', 'isfinite', 'isinf', 'isnan',
              'left_shift', 'less', 'less_equal', 'log', 'log10', 'log1p',
              'log2', 'logaddexp', 'logaddexp2', 'logical_and', 'logical_not',
              'logical_or', 'logical_xor', 'maximum', 'minimum', 'mod', 'modf',
              'multiply', 'negative', 'not_equal', 'power',
              'rad2deg', 'reciprocal', 'remainder', 'right_shift', 'rint',
              'sign', 'signbit', 'sin', 'sinh', 'sqrt', 'square', 'subtract',
              'tan', 'tanh', 'true_divide', 'trunc']
# ,'isreal', 'iscomplex', 'ldexp', 'frexp'

# Add some standardized information
UFUNCS = []
for name in RAW_UFUNCS:
    ufunc = getattr(np, name)
    n_in, n_out = ufunc.nin, ufunc.nout
    descr = ufunc.__doc__.splitlines()[2]
    # Numpy occasionally uses single ticks for doc, we only use them for links
    descr = re.sub('`+', '``', descr)
    doc = descr + """

See Also
--------
numpy.{}
""".format(name)
    UFUNCS.append((name, n_in, n_out, doc))

# TODO: add the following reductions (to the CUDA implementation):
# ['var', 'trace', 'tensordot', 'std', 'ptp', 'mean', 'diff', 'cumsum',
#  'cumprod', 'average']


# --- Wrappers for `Tensor` --- #


def wrap_ufunc_base(name, n_in, n_out, doc):
    """Return ufunc wrapper for implementation-agnostic ufunc classes."""
    ufunc = getattr(np, name)
    if n_in == 1:
        if n_out == 1:
            def wrapper(self, out=None, **kwargs):
                if out is None or isinstance(out, (type(self.elem),
                                                   type(self.elem.data))):
                    out = (out,)

                return self.elem.__array_ufunc__(
                    ufunc, '__call__', self.elem, out=out, **kwargs)

        elif n_out == 2:
            def wrapper(self, out=None, **kwargs):
                if out is None:
                    out = (None, None)

                return self.elem.__array_ufunc__(
                    ufunc, '__call__', self.elem, out=out, **kwargs)

        else:
            raise NotImplementedError

    elif n_in == 2:
        if n_out == 1:
            def wrapper(self, x2, out=None, **kwargs):
                return self.elem.__array_ufunc__(
                    ufunc, '__call__', self.elem, x2, out=(out,), **kwargs)

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    wrapper.__name__ = wrapper.__qualname__ = name
    wrapper.__doc__ = doc
    return wrapper


class TensorSpaceUfuncs(object):

    """Ufuncs for `Tensor` objects.

    Internal object, should not be created except in `Tensor`.
    """

    def __init__(self, elem):
        """Create ufunc wrapper for elem."""
        self.elem = elem

    # Reductions for backwards compatibility

    def sum(self, axis=None, dtype=None, out=None, keepdims=False):
        """Return the sum of ``self``.

        See Also
        --------
        numpy.sum
        prod
        """
        return self.elem.__array_ufunc__(
            np.add, 'reduce', self.elem,
            axis=axis, dtype=dtype, out=(out,), keepdims=keepdims)

    def prod(self, axis=None, dtype=None, out=None, keepdims=False):
        """Return the product of ``self``.

        See Also
        --------
        numpy.prod
        sum
        """
        return self.elem.__array_ufunc__(
            np.multiply, 'reduce', self.elem,
            axis=axis, dtype=dtype, out=(out,), keepdims=keepdims)

    def min(self, axis=None, dtype=None, out=None, keepdims=False):
        """Return the minimum of ``self``.

        See Also
        --------
        numpy.amin
        max
        """
        return self.elem.__array_ufunc__(
            np.minimum, 'reduce', self.elem,
            axis=axis, dtype=dtype, out=(out,), keepdims=keepdims)

    def max(self, axis=None, dtype=None, out=None, keepdims=False):
        """Return the maximum of ``self``.

        See Also
        --------
        numpy.amax
        min
        """
        return self.elem.__array_ufunc__(
            np.maximum, 'reduce', self.elem,
            axis=axis, dtype=dtype, out=(out,), keepdims=keepdims)


# Add ufunc methods to ufunc class
for name, n_in, n_out, doc in UFUNCS:
    method = wrap_ufunc_base(name, n_in, n_out, doc)
    setattr(TensorSpaceUfuncs, name, method)


# --- Wrappers for `ProductSpaceElement` --- #


def wrap_ufunc_productspace(name, n_in, n_out, doc):
    """Return ufunc wrapper for `ProductSpaceUfuncs`."""
    if n_in == 1:
        if n_out == 1:
            def wrapper(self, out=None, **kwargs):
                if out is None:
                    result = [getattr(x.ufuncs, name)(**kwargs)
                              for x in self.elem]
                    return self.elem.space.element(result)
                else:
                    for x, out_x in zip(self.elem, out):
                        getattr(x.ufuncs, name)(out=out_x, **kwargs)
                    return out

        elif n_out == 2:
            def wrapper(self, out1=None, out2=None, **kwargs):
                if out1 is None:
                    out1 = self.elem.space.element()
                if out2 is None:
                    out2 = self.elem.space.element()
                for x, out1_x, out2_x in zip(self.elem, out1, out2):
                    getattr(x.ufuncs, name)(out1=out1_x, out2=out2_x, **kwargs)
                return out1, out2

        else:
            raise NotImplementedError

    elif n_in == 2:
        if n_out == 1:
            def wrapper(self, x2, out=None, **kwargs):
                if x2 in self.elem.space:
                    if out is None:
                        result = [getattr(x.ufuncs, name)(x2p, **kwargs)
                                  for x, x2p in zip(self.elem, x2)]
                        return self.elem.space.element(result)
                    else:
                        for x, x2p, outp in zip(self.elem, x2, out):
                            getattr(x.ufuncs, name)(x2p, out=outp, **kwargs)
                        return out
                else:
                    if out is None:
                        result = [getattr(x.ufuncs, name)(x2, **kwargs)
                                  for x in self.elem]
                        return self.elem.space.element(result)
                    else:
                        for x, outp in zip(self.elem, out):
                            getattr(x.ufuncs, name)(x2, out=outp, **kwargs)
                        return out

        else:
            raise NotImplementedError
    else:
        raise NotImplementedError

    wrapper.__name__ = wrapper.__qualname__ = name
    wrapper.__doc__ = doc
    return wrapper


class ProductSpaceUfuncs(object):

    """Ufuncs for `ProductSpaceElement` objects.

    Internal object, should not be created except in `ProductSpaceElement`.
    """
    def __init__(self, elem):
        """Create ufunc wrapper for ``elem``."""
        self.elem = elem

    def sum(self):
        """Return the sum of ``self``.

        See Also
        --------
        numpy.sum
        prod
        """
        results = [x.ufuncs.sum() for x in self.elem]
        return np.sum(results)

    def prod(self):
        """Return the product of ``self``.

        See Also
        --------
        numpy.prod
        sum
        """
        results = [x.ufuncs.prod() for x in self.elem]
        return np.prod(results)

    def min(self):
        """Return the minimum of ``self``.

        See Also
        --------
        numpy.amin
        max
        """
        results = [x.ufuncs.min() for x in self.elem]
        return np.min(results)

    def max(self):
        """Return the maximum of ``self``.

        See Also
        --------
        numpy.amax
        min
        """
        results = [x.ufuncs.max() for x in self.elem]
        return np.max(results)


# Add ufunc methods to ufunc class
for name, n_in, n_out, doc in UFUNCS:
    method = wrap_ufunc_productspace(name, n_in, n_out, doc)
    setattr(ProductSpaceUfuncs, name, method)



# coding: utf-8
# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import print_function, division, absolute_import
import numpy as np

from odl.discr import ResizingOperator
from odl.trafos import FourierTransform, PYFFTW_AVAILABLE


__all__ = ('fbp_op', 'fbp_filter_op', 'tam_danielson_window',
           'parker_weighting')


def _axis_in_detector(geometry):
    """A vector in the detector plane that points along the rotation axis."""
    du, dv = geometry.det_axes_init
    axis = geometry.axis
    c = np.array([np.vdot(axis, du), np.vdot(axis, dv)])
    cnorm = np.linalg.norm(c)

    # Check for numerical errors
    assert cnorm != 0

    return c / cnorm


def _rotation_direction_in_detector(geometry):
    """A vector in the detector plane that points in the rotation direction."""
    du, dv = geometry.det_axes_init
    axis = geometry.axis
    det_normal = np.cross(dv, du)
    rot_dir = np.cross(axis, det_normal)
    c = np.array([np.vdot(rot_dir, du), np.vdot(rot_dir, dv)])
    cnorm = np.linalg.norm(c)

    # Check for numerical errors
    assert cnorm != 0

    return c / cnorm


def _fbp_filter(norm_freq, filter_type, frequency_scaling):
    """Create a smoothing filter for FBP.

    Parameters
    ----------
    norm_freq : `array-like`
        Frequencies normalized to lie in the interval [0, 1].
    filter_type : {'Ram-Lak', 'Shepp-Logan', 'Cosine', 'Hamming', 'Hann',
                   callable}
        The type of filter to be used.
        If a string is given, use one of the standard filters with that name.
        A callable should take an array of values in [0, 1] and return the
        filter for these frequencies.
    frequency_scaling : float
        Scaling of the frequencies for the filter. All frequencies are scaled
        by this number, any relative frequency above ``frequency_scaling`` is
        set to 0.

    Returns
    -------
    smoothing_filter : `numpy.ndarray`

    Examples
    --------
    Create an FBP filter

    >>> norm_freq = np.linspace(0, 1, 10)
    >>> filt = _fbp_filter(norm_freq,
    ...                    filter_type='Hann',
    ...                    frequency_scaling=0.8)
    """
    filter_type, filter_type_in = str(filter_type).lower(), filter_type
    if callable(filter_type):
        filt = filter_type(norm_freq)
    elif filter_type == 'ram-lak':
        filt = np.copy(norm_freq)
    elif filter_type == 'shepp-logan':
        filt = norm_freq * np.sinc(norm_freq / (2 * frequency_scaling))
    elif filter_type == 'cosine':
        filt = norm_freq * np.cos(norm_freq * np.pi / (2 * frequency_scaling))
    elif filter_type == 'hamming':
        filt = norm_freq * (
            0.54 + 0.46 * np.cos(norm_freq * np.pi / (frequency_scaling)))
    elif filter_type == 'hann':
        filt = norm_freq * (
            np.cos(norm_freq * np.pi / (2 * frequency_scaling)) ** 2)
    else:
        raise ValueError('unknown `filter_type` ({})'
                         ''.format(filter_type_in))

    indicator = (norm_freq <= frequency_scaling)
    filt *= indicator
    return filt


def tam_danielson_window(ray_trafo, smoothing_width=0.05, n_pi=1):
    """Create Tam-Danielson window from a `RayTransform`.

    The Tam-Danielson window is an indicator function on the minimal set of
    data needed to reconstruct a volume from given data. It is useful in
    analytic reconstruction methods such as FBP to give a more accurate
    reconstruction.

    See [TAM1998] for more informationon the window.
    See [PKGT2000] for information on the ``n_pi`` parameter.

    Parameters
    ----------
    ray_trafo : `RayTransform`
        The ray transform for which to compute the window.
    smoothing_width : positive float, optional
        Width of the smoothing applied to the window's edges given as a
        fraction of the width of the full window.
    n_pi : odd int, optional
        Total number of half rotations to include in the window. Values larger
        than 1 should be used if the pitch is much smaller than the detector
        height.

    Returns
    -------
    tam_danielson_window : ``ray_trafo.range`` element

    See Also
    --------
    fbp_op : Filtered back-projection operator from `RayTransform`
    tam_danielson_window : Weighting for short scan data
    odl.tomo.geometry.conebeam.ConeBeamGeometry :
        Primary use case for this window function.

    References
    ----------
    [TSS1998] Tam, K C, Samarasekera, S and Sauer, F.
    *Exact cone beam CT with a spiral scan*.
    Physics in Medicine & Biology 4 (1998), p 1015.
    https://dx.doi.org/10.1088/0031-9155/43/4/028

    [PKGT2000] Proksa R, Köhler T, Grass M, Timmer J.
    *The n-PI-method for helical cone-beam CT*
    IEEE Trans Med Imaging. 2000 Sep;19(9):848-63.
    https://www.ncbi.nlm.nih.gov/pubmed/11127600
    """
    # Extract parameters
    src_radius = ray_trafo.geometry.src_radius
    det_radius = ray_trafo.geometry.det_radius
    pitch = ray_trafo.geometry.pitch

    if pitch == 0:
        raise ValueError('Tam-Danielson window is only defined with '
                         '`pitch!=0`')

    smoothing_width = float(smoothing_width)
    if smoothing_width < 0:
        raise ValueError('`smoothing_width` should be a positive float')

    if n_pi % 2 != 1:
        raise ValueError('`n_pi` must be odd, got {}'.format(n_pi))

    # Find projection of axis on detector
    axis_proj = _axis_in_detector(ray_trafo.geometry)
    rot_dir = _rotation_direction_in_detector(ray_trafo.geometry)

    # Find distance from projection of rotation axis for each pixel
    dx = (rot_dir[0] * ray_trafo.range.meshgrid[1]
          + rot_dir[1] * ray_trafo.range.meshgrid[2])

    dx_axis = dx * src_radius / (src_radius + det_radius)

    def Vn(u):
        return (pitch / (2 * np.pi)
                * (1 + (u / src_radius) ** 2)
                * (n_pi * np.pi / 2.0 - np.arctan(u / src_radius)))

    lower_proj_axis = -Vn(dx_axis)
    upper_proj_axis = Vn(-dx_axis)

    lower_proj = lower_proj_axis * (src_radius + det_radius) / src_radius
    upper_proj = upper_proj_axis * (src_radius + det_radius) / src_radius

    # Compute a smoothed width
    interval = (upper_proj - lower_proj)
    width = interval * smoothing_width / np.sqrt(2)

    # Create window function
    def window_fcn(x):
        # Lazy import to improve `import odl` time
        import scipy.special

        x_along_axis = axis_proj[0] * x[1] + axis_proj[1] * x[2]
        if smoothing_width != 0:
            lower_wndw = 0.5 * (
                1 + scipy.special.erf((x_along_axis - lower_proj) / width))
            upper_wndw = 0.5 * (
                1 + scipy.special.erf((upper_proj - x_along_axis) / width))
        else:
            lower_wndw = (x_along_axis >= lower_proj)
            upper_wndw = (x_along_axis <= upper_proj)

        return lower_wndw * upper_wndw

    return ray_trafo.range.element(window_fcn) / n_pi


def parker_weighting(ray_trafo, q=0.25):
    """Create parker weighting for a `RayTransform`.

    Parker weighting is a weighting function that ensures that oversampled
    fan/cone beam data are weighted such that each line has unit weight. It is
    useful in analytic reconstruction methods such as FBP to give a more
    accurate result and can improve convergence rates for iterative methods.

    See the article `Parker weights revisited`_ for more information.

    Parameters
    ----------
    ray_trafo : `RayTransform`
        The ray transform for which to compute the weights.
    q : float, optional
        Parameter controlling the speed of the roll-off at the edges of the
        weighting. 1.0 gives the classical Parker weighting, while smaller
        values in general lead to lower noise but stronger discretization
        artifacts.

    Returns
    -------
    parker_weighting : ``ray_trafo.range`` element

    See Also
    --------
    fbp_op : Filtered back-projection operator from `RayTransform`
    tam_danielson_window : Indicator function for helical data
    odl.tomo.geometry.conebeam.FanBeamGeometry : Use case in 2d
    odl.tomo.geometry.conebeam.ConeBeamGeometry : Use case in 3d (for pitch 0)

    References
    ----------
    .. _Parker weights revisited: https://www.ncbi.nlm.nih.gov/pubmed/11929021
    """
    # Note: Parameter names taken from WES2002

    # Extract parameters
    src_radius = ray_trafo.geometry.src_radius
    det_radius = ray_trafo.geometry.det_radius
    ndim = ray_trafo.geometry.ndim
    angles = ray_trafo.range.meshgrid[0]
    min_rot_angle = ray_trafo.geometry.motion_partition.min_pt
    alen = ray_trafo.geometry.motion_params.length

    # Parker weightings are not defined for helical geometries
    if ray_trafo.geometry.ndim != 2:
        pitch = ray_trafo.geometry.pitch
        if pitch != 0:
            raise ValueError('Parker weighting window is only defined with '
                             '`pitch==0`')

    # Find distance from projection of rotation axis for each pixel
    if ndim == 2:
        dx = ray_trafo.range.meshgrid[1]
    elif ndim == 3:
        # Find projection of axis on detector
        rot_dir = _rotation_direction_in_detector(ray_trafo.geometry)
        # If axis is aligned to a coordinate axis, save some memory and time by
        # using broadcasting
        if rot_dir[0] == 0:
            dx = rot_dir[1] * ray_trafo.range.meshgrid[2]
        elif rot_dir[1] == 0:
            dx = rot_dir[0] * ray_trafo.range.meshgrid[1]
        else:
            dx = (rot_dir[0] * ray_trafo.range.meshgrid[1]
                  + rot_dir[1] * ray_trafo.range.meshgrid[2])

    # Compute parameters
    dx_abs_max = np.max(np.abs(dx))
    max_fan_angle = 2 * np.arctan2(dx_abs_max, src_radius + det_radius)
    delta = max_fan_angle / 2
    epsilon = alen - np.pi - max_fan_angle

    if epsilon < 0:
        raise Exception('data not sufficiently sampled for parker weighting')

    # Define utility functions
    def S(betap):
        return (0.5 * (1.0 + np.sin(np.pi * betap)) * (np.abs(betap) < 0.5)
                + (betap >= 0.5))

    def b(alpha):
        return q * (2 * delta - 2 * alpha + epsilon)

    # Create weighting function
    beta = np.asarray(angles - min_rot_angle,
                      dtype=ray_trafo.range.dtype)  # rotation angle
    alpha = np.asarray(np.arctan2(dx, src_radius + det_radius),
                       dtype=ray_trafo.range.dtype)

    # Compute sum in place to save memory
    S_sum = S(beta / b(alpha) - 0.5)
    S_sum += S((beta - 2 * delta + 2 * alpha - epsilon) / b(alpha) + 0.5)
    S_sum -= S((beta - np.pi + 2 * alpha) / b(-alpha) - 0.5)
    S_sum -= S((beta - np.pi - 2 * delta - epsilon) / b(-alpha) + 0.5)

    scale = 0.5 * alen / np.pi
    return ray_trafo.range.element(
        np.broadcast_to(S_sum * scale, ray_trafo.range.shape))


def fbp_filter_op(ray_trafo, padding=True, filter_type='Ram-Lak',
                  frequency_scaling=1.0):
    """Create a filter operator for FBP from a `RayTransform`.

    Parameters
    ----------
    ray_trafo : `RayTransform`
        The ray transform (forward operator) whose approximate inverse should
        be computed. Its geometry has to be any of the following

        `Parallel2dGeometry` : Exact reconstruction

        `Parallel3dAxisGeometry` : Exact reconstruction

        `FanBeamGeometry` : Approximate reconstruction, correct in limit of
        fan angle = 0.
        Only flat detectors are supported (det_curvature_radius is None).

        `ConeBeamGeometry`, pitch = 0 (circular) : Approximate reconstruction,
        correct in the limit of fan angle = 0 and cone angle = 0.

        `ConeBeamGeometry`, pitch > 0 (helical) : Very approximate unless a
        `tam_danielson_window` is used. Accurate with the window.

        Other geometries: Not supported

    padding : bool, optional
        If the data space should be zero padded. Without padding, the data may
        be corrupted due to the circular convolution used. Using padding makes
        the algorithm slower.
    filter_type : optional
        The type of filter to be used.
        The predefined options are, in approximate order from most noise
        senstive to least noise sensitive:
        ``'Ram-Lak'``, ``'Shepp-Logan'``, ``'Cosine'``, ``'Hamming'`` and
        ``'Hann'``.
        A callable can also be provided. It must take an array of values in
        [0, 1] and return the filter for these frequencies.
    frequency_scaling : float, optional
        Relative cutoff frequency for the filter.
        The normalized frequencies are rescaled so that they fit into the range
        [0, frequency_scaling]. Any frequency above ``frequency_scaling`` is
        set to zero.

    Returns
    -------
    filter_op : `Operator`
        Filtering operator for FBP based on ``ray_trafo``.

    See Also
    --------
    tam_danielson_window : Windowing for helical data
    """
    impl = 'pyfftw' if PYFFTW_AVAILABLE else 'numpy'
    alen = ray_trafo.geometry.motion_params.length

    if ray_trafo.domain.ndim == 2:
        # Define ramp filter
        def fourier_filter(x):
            abs_freq = np.abs(x[1])
            norm_freq = abs_freq / np.max(abs_freq)
            filt = _fbp_filter(norm_freq, filter_type, frequency_scaling)
            scaling = 1 / (2 * alen)
            return filt * np.max(abs_freq) * scaling

        # Define (padded) fourier transform
        if padding:
            # Define padding operator
            ran_shp = (ray_trafo.range.shape[0],
                       ray_trafo.range.shape[1] * 2 - 1)
            resizing = ResizingOperator(ray_trafo.range, ran_shp=ran_shp)

            fourier = FourierTransform(resizing.range, axes=1, impl=impl)
            fourier = fourier * resizing
        else:
            fourier = FourierTransform(ray_trafo.range, axes=1, impl=impl)

    elif ray_trafo.domain.ndim == 3:
        # Find the direction that the filter should be taken in
        rot_dir = _rotation_direction_in_detector(ray_trafo.geometry)

        # Find what axes should be used in the fourier transform
        used_axes = (rot_dir != 0)
        if used_axes[0] and not used_axes[1]:
            axes = [1]
        elif not used_axes[0] and used_axes[1]:
            axes = [2]
        else:
            axes = [1, 2]

        # Add scaling for cone-beam case
        if hasattr(ray_trafo.geometry, 'src_radius'):
            scale = (ray_trafo.geometry.src_radius
                     / (ray_trafo.geometry.src_radius
                        + ray_trafo.geometry.det_radius))

            if ray_trafo.geometry.pitch != 0:
                # In helical geometry the whole volume is not in each
                # projection and we need to use another weighting.
                # Ideally each point in the volume effects only
                # the projections in a half rotation, so we assume that that
                # is the case.
                scale *= alen / (np.pi)
        else:
            scale = 1.0

        # Define ramp filter
        def fourier_filter(x):
            # If axis is aligned to a coordinate axis, save some memory and
            # time by using broadcasting
            if not used_axes[0]:
                abs_freq = np.abs(rot_dir[1] * x[2])
            elif not used_axes[1]:
                abs_freq = np.abs(rot_dir[0] * x[1])
            else:
                abs_freq = np.abs(rot_dir[0] * x[1] + rot_dir[1] * x[2])
            norm_freq = abs_freq / np.max(abs_freq)
            filt = _fbp_filter(norm_freq, filter_type, frequency_scaling)
            scaling = scale * np.max(abs_freq) / (2 * alen)
            return filt * scaling

        # Define (padded) fourier transform
        if padding:
            # Define padding operator
            if used_axes[0]:
                padded_shape_u = ray_trafo.range.shape[1] * 2 - 1
            else:
                padded_shape_u = ray_trafo.range.shape[1]

            if used_axes[1]:
                padded_shape_v = ray_trafo.range.shape[2] * 2 - 1
            else:
                padded_shape_v = ray_trafo.range.shape[2]

            ran_shp = (ray_trafo.range.shape[0],
                       padded_shape_u,
                       padded_shape_v)
            resizing = ResizingOperator(ray_trafo.range, ran_shp=ran_shp)

            fourier = FourierTransform(resizing.range, axes=axes, impl=impl)
            fourier = fourier * resizing
        else:
            fourier = FourierTransform(ray_trafo.range, axes=axes, impl=impl)
    else:
        raise NotImplementedError('FBP only implemented in 2d and 3d')

    # Create ramp in the detector direction
    ramp_function = fourier.range.element(fourier_filter)

    weight = 1
    if not ray_trafo.range.is_weighted:
        # Compensate for potentially unweighted range of the ray transform
        weight *= ray_trafo.range.cell_volume

    if not ray_trafo.domain.is_weighted:
        # Compensate for potentially unweighted domain of the ray transform
        weight /= ray_trafo.domain.cell_volume

    ramp_function *= weight

    # Create ramp filter via the convolution formula with fourier transforms
    return fourier.inverse * ramp_function * fourier


def fbp_op(ray_trafo, padding=True, filter_type='Ram-Lak',
           frequency_scaling=1.0):
    """Create filtered back-projection operator from a `RayTransform`.

    The filtered back-projection is an approximate inverse to the ray
    transform.

    Parameters
    ----------
    ray_trafo : `RayTransform`
        The ray transform (forward operator) whose approximate inverse should
        be computed. Its geometry has to be any of the following

        `Parallel2dGeometry` : Exact reconstruction

        `Parallel3dAxisGeometry` : Exact reconstruction

        `FanBeamGeometry` : Approximate reconstruction, correct in limit of fan
        angle = 0.
        Only flat detectors are supported (det_curvature_radius is None).

        `ConeBeamGeometry`, pitch = 0 (circular) : Approximate reconstruction,
        correct in the limit of fan angle = 0 and cone angle = 0.

        `ConeBeamGeometry`, pitch > 0 (helical) : Very approximate unless a
        `tam_danielson_window` is used. Accurate with the window.

        Other geometries: Not supported

    padding : bool, optional
        If the data space should be zero padded. Without padding, the data may
        be corrupted due to the circular convolution used. Using padding makes
        the algorithm slower.
    filter_type : optional
        The type of filter to be used.
        The predefined options are, in approximate order from most noise
        senstive to least noise sensitive:
        ``'Ram-Lak'``, ``'Shepp-Logan'``, ``'Cosine'``, ``'Hamming'`` and
        ``'Hann'``.
        A callable can also be provided. It must take an array of values in
        [0, 1] and return the filter for these frequencies.
    frequency_scaling : float, optional
        Relative cutoff frequency for the filter.
        The normalized frequencies are rescaled so that they fit into the range
        [0, frequency_scaling]. Any frequency above ``frequency_scaling`` is
        set to zero.

    Returns
    -------
    fbp_op : `Operator`
        Approximate inverse operator of ``ray_trafo``.

    See Also
    --------
    tam_danielson_window : Windowing for helical data.
    parker_weighting : Windowing for overcomplete fan-beam data.
    """
    return ray_trafo.adjoint * fbp_filter_op(ray_trafo, padding, filter_type,
                                             frequency_scaling)


if __name__ == '__main__':
    import odl
    import matplotlib.pyplot as plt
    from odl.util.testutils import run_doctests

    # Display the various filters
    x = np.linspace(0, 1, 100)
    cutoff = 0.7

    plt.figure('fbp filter')
    for filter_name in ['Ram-Lak', 'Shepp-Logan', 'Cosine', 'Hamming', 'Hann',
                        np.sqrt]:
        plt.plot(x, _fbp_filter(x, filter_name, cutoff), label=filter_name)

    plt.title('Filters with frequency scaling = {}'.format(cutoff))
    plt.legend(loc=2)

    # Show the Tam-Danielson window

    # Create Ray Transform in helical geometry
    reco_space = odl.uniform_discr(
        min_pt=[-20, -20, 0], max_pt=[20, 20, 40], shape=[300, 300, 300])
    angle_partition = odl.uniform_partition(0, 8 * 2 * np.pi, 2000)
    detector_partition = odl.uniform_partition([-40, -4], [40, 4], [500, 500])
    geometry = odl.tomo.ConeBeamGeometry(
        angle_partition, detector_partition, src_radius=100, det_radius=100,
        pitch=5.0)
    ray_trafo = odl.tomo.RayTransform(reco_space, geometry, impl='astra_cuda')

    # Crete and show TD window
    td_window = tam_danielson_window(ray_trafo, smoothing_width=0)
    td_window.show('Tam-Danielson window', coords=[0, None, None])

    # Show the Parker weighting

    # Create Ray Transform in fan beam geometry
    geometry = odl.tomo.cone_beam_geometry(reco_space,
                                           src_radius=40, det_radius=80)
    ray_trafo = odl.tomo.RayTransform(reco_space, geometry, impl='astra_cuda')

    # Crete and show parker weighting
    parker_weighting = parker_weighting(ray_trafo)
    parker_weighting.show('Parker weighting')

    # Also run the doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for converting ODL spaces to tensorflow layers."""

from __future__ import print_function, division, absolute_import
import tensorflow as tf
import numpy as np

import odl


__all__ = ('TensorflowOperator',)


class TensorflowOperator(odl.Operator):
    def __init__(self, input_tensor, output_tensor,
                 domain=None, range=None,
                 linear=False, sess=None):
        """Wrap Tensorflow layers in ODL operator.

        Parameters
        ----------
        input_tensor : `tf.Tensor`
            Input to the tensorflow graph (values will be fed to this).
        output_tensor : `tf.Tensor`
            Output node from the graph.
        domain : `TensorSpace`, optional
            Domain of the wrapping operator.
            Default: `tensor_space` with same shape and dtype as
            ``input_tensor``.
        range : `TensorSpace`, optional
            Range of the wrapping operator.
            Default: `tensor_space` with same shape and dtype as
            ``output_tensor``.
        linear : bool, optional
            If the created operator should be linear
        sess : `tf.Session`, optional
            Session to evaluate the graph in.
            Default: `tf.get_default_session`

        Notes
        -----
        Both `Operator.derivative` and `Operator.adjoint` are implemented
        using automatic differentiation in tensorflow.
        """
        self.input_tensor = input_tensor
        self.output_tensor = output_tensor

        if domain is None:
            domain = odl.tensor_space(input_tensor.shape.as_list(),
                                      dtype=input_tensor.dtype.as_numpy_dtype)
        if range is None:
            range = odl.tensor_space(output_tensor.shape.as_list(),
                                     dtype=output_tensor.dtype.as_numpy_dtype)

        # TODO: replace with #1177
        self.adjoint_weight = domain.weighting.const / range.weighting.const

        self.dx = tf.placeholder(input_tensor.dtype,
                                 shape=input_tensor.shape)
        self.dy = tf.placeholder(output_tensor.dtype,
                                 shape=output_tensor.shape)

        adjoint_of_derivative_tensor = tf.gradients(
            self.output_tensor, [self.input_tensor], [self.dy])[0]
        self.adjoint_of_derivative_tensor = (range.weighting.const *
                                             adjoint_of_derivative_tensor)

        # Since tensorflow does not support forward differentiation, use trick
        # that adjoint of the derivative of adjoint of the derivative is simply
        # the derivative.
        derivative_tensor = tf.gradients(
            adjoint_of_derivative_tensor, [self.dy], [self.dx])[0]
        self.derivative_tensor = (range.weighting.const *
                                  derivative_tensor)

        if sess is None:
            self.sess = tf.get_default_session()
        else:
            self.sess = sess

        super(TensorflowOperator, self).__init__(domain, range, linear=linear)

    def _call(self, x):
        """Return ``self(x)``."""
        result = self.sess.run(self.output_tensor,
                               feed_dict={self.input_tensor: np.asarray(x)})

        return result

    def derivative(self, x):
        op = self

        class TensorflowOperatorDerivative(odl.Operator):
            def _call(self, dx):
                result = op.sess.run(op.derivative_tensor,
                                     feed_dict={op.input_tensor: np.asarray(x),
                                                op.dx: np.asarray(dx)})

                return result

            @property
            def adjoint(self):
                class TensorflowOperatorDerivativeAdjoint(odl.Operator):
                    def _call(self, y):
                        result = op.sess.run(
                            op.adjoint_of_derivative_tensor,
                            feed_dict={op.input_tensor: np.asarray(x),
                                       op.dy: np.asarray(y)})

                        # Compensate for weighted spaces in ODL
                        result *= op.adjoint_weight
                        return result

                return TensorflowOperatorDerivativeAdjoint(self.range,
                                                           self.domain,
                                                           linear=True)

        return TensorflowOperatorDerivative(self.domain,
                                            self.range,
                                            linear=True)


if __name__ == '__main__':
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tomographic datasets from Mayo Clinic.

In addition to the standard ODL requirements, this library also requires:

    - tqdm
    - dicom
    - A copy of the Mayo dataset, see
    https://www.aapm.org/GrandChallenge/LowDoseCT/#registration
"""

from __future__ import division
import numpy as np
import os
import dicom
import odl
import tqdm

from dicom.datadict import DicomDictionary, NameDict, CleanName
from odl.discr.discr_utils import linear_interpolator
from odl.contrib.datasets.ct.mayo_dicom_dict import new_dict_items

# Update the DICOM dictionary with the extra Mayo tags
DicomDictionary.update(new_dict_items)
NameDict.update((CleanName(tag), tag) for tag in new_dict_items)


__all__ = ('load_projections', 'load_reconstruction')


def _read_projections(folder, indices):
    """Read mayo projections from a folder."""
    datasets = []

    # Get the relevant file names
    file_names = sorted([f for f in os.listdir(folder) if f.endswith(".dcm")])

    if len(file_names) == 0:
        raise ValueError('No DICOM files found in {}'.format(folder))

    file_names = file_names[indices]

    data_array = None

    for i, file_name in enumerate(tqdm.tqdm(file_names,
                                            'Loading projection data')):
        # read the file
        dataset = dicom.read_file(folder + '/' + file_name)

        # Get some required data
        rows = dataset.NumberofDetectorRows
        cols = dataset.NumberofDetectorColumns
        hu_factor = dataset.HUCalibrationFactor
        rescale_intercept = dataset.RescaleIntercept
        rescale_slope = dataset.RescaleSlope

        # Load the array as bytes
        proj_array = np.array(np.frombuffer(dataset.PixelData, 'H'),
                              dtype='float32')
        proj_array = proj_array.reshape([rows, cols], order='F').T

        # Rescale array
        proj_array *= rescale_slope
        proj_array += rescale_intercept
        proj_array /= hu_factor

        # Store results
        if data_array is None:
            # We need to load the first dataset before we know the shape
            data_array = np.empty((len(file_names), cols, rows),
                                  dtype='float32')

        data_array[i] = proj_array[:, ::-1]
        datasets.append(dataset)

    return datasets, data_array


def load_projections(folder, indices=None):
    """Load geometry and data stored in Mayo format from folder.

    Parameters
    ----------
    folder : str
        Path to the folder where the Mayo DICOM files are stored.
    indices : optional
        Indices of the projections to load.
        Accepts advanced indexing such as slice or list of indices.

    Returns
    -------
    geometry : ConeBeamGeometry
        Geometry corresponding to the Mayo projector.
    proj_data : `numpy.ndarray`
        Projection data, given as the line integral of the linear attenuation
        coefficient (g/cm^3). Its unit is thus g/cm^2.
    """
    datasets, data_array = _read_projections(folder, indices)

    # Get the angles
    angles = [d.DetectorFocalCenterAngularPosition for d in datasets]
    angles = -np.unwrap(angles) - np.pi  # different definition of angles

    # Set minimum and maximum corners
    shape = np.array([datasets[0].NumberofDetectorColumns,
                      datasets[0].NumberofDetectorRows])
    pixel_size = np.array([datasets[0].DetectorElementTransverseSpacing,
                           datasets[0].DetectorElementAxialSpacing])

    # Correct from center of pixel to corner of pixel
    minp = -(np.array(datasets[0].DetectorCentralElement) - 0.5) * pixel_size
    maxp = minp + shape * pixel_size

    # Select geometry parameters
    src_radius = datasets[0].DetectorFocalCenterRadialDistance
    det_radius = (datasets[0].ConstantRadialDistance -
                  datasets[0].DetectorFocalCenterRadialDistance)

    # For unknown reasons, mayo does not include the tag
    # "TableFeedPerRotation", which is what we want.
    # Instead we manually compute the pitch
    pitch = ((datasets[-1].DetectorFocalCenterAxialPosition -
              datasets[0].DetectorFocalCenterAxialPosition) /
             ((np.max(angles) - np.min(angles)) / (2 * np.pi)))

    # Get flying focal spot data
    offset_axial = np.array([d.SourceAxialPositionShift for d in datasets])
    offset_angular = np.array([d.SourceAngularPositionShift for d in datasets])
    offset_radial = np.array([d.SourceRadialDistanceShift for d in datasets])

    # TODO(adler-j): Implement proper handling of flying focal spot.
    # Currently we do not fully account for it, merely making some "first
    # order corrections" to the detector position and radial offset.

    # Update angles with flying focal spot (in plane direction).
    # This increases the resolution of the reconstructions.
    angles = angles - offset_angular

    # We correct for the mean offset due to the rotated angles, we need to
    # shift the detector.
    offset_detector_by_angles = det_radius * np.mean(offset_angular)
    minp[0] -= offset_detector_by_angles
    maxp[0] -= offset_detector_by_angles

    # We currently apply only the mean of the offsets
    src_radius = src_radius + np.mean(offset_radial)

    # Partially compensate for a movement of the source by moving the object
    # instead. We need to rescale by the magnification to get the correct
    # change in the detector. This approximation is only exactly valid on the
    # axis of rotation.
    mean_offset_along_axis_for_ffz = np.mean(offset_axial) * (
        src_radius / (src_radius + det_radius))

    # Create partition for detector
    detector_partition = odl.uniform_partition(minp, maxp, shape)

    # Convert offset to odl definitions
    offset_along_axis = (mean_offset_along_axis_for_ffz +
                         datasets[0].DetectorFocalCenterAxialPosition -
                         angles[0] / (2 * np.pi) * pitch)

    # Assemble geometry
    angle_partition = odl.nonuniform_partition(angles)
    geometry = odl.tomo.ConeBeamGeometry(angle_partition,
                                         detector_partition,
                                         src_radius=src_radius,
                                         det_radius=det_radius,
                                         pitch=pitch,
                                         offset_along_axis=offset_along_axis)

    # Create a *temporary* ray transform (we need its range)
    spc = odl.uniform_discr([-1] * 3, [1] * 3, [32] * 3)
    ray_trafo = odl.tomo.RayTransform(spc, geometry, interp='linear')

    # convert coordinates
    theta, up, vp = ray_trafo.range.grid.meshgrid
    d = src_radius + det_radius
    u = d * np.arctan(up / d)
    v = d / np.sqrt(d**2 + up**2) * vp

    # Calculate projection data in rectangular coordinates since we have no
    # backend that supports cylindrical
    interpolator = linear_interpolator(
        data_array, ray_trafo.range.coord_vectors
    )
    proj_data = interpolator((theta, u, v))

    return geometry, proj_data.asarray()


def load_reconstruction(folder, slice_start=0, slice_end=-1):
    """Load a volume from folder, also returns the corresponding partition.

    Parameters
    ----------
    folder : str
        Path to the folder where the DICOM files are stored.
    slice_start : int
        Index of the first slice to use. Used for subsampling.
    slice_end : int
        Index of the final slice to use.

    Returns
    -------
    partition : `odl.RectPartition`
        Partition describing the geometric positioning of the voxels.
    data : `numpy.ndarray`
        Volumetric data. Scaled such that data = 1 for water (0 HU).

    Notes
    -----
    DICOM data is highly non trivial. Typically, each slice has been computed
    with a slice tickness (e.g. 3mm) but the slice spacing might be
    different from that.

    Further, the coordinates in DICOM is typically the *middle* of the pixel,
    not the corners as in ODL.

    This function should handle all of these peculiarities and give a volume
    with the correct coordinate system attached.
    """
    file_names = sorted([f for f in os.listdir(folder) if f.endswith(".IMA")])

    if len(file_names) == 0:
        raise ValueError('No DICOM files found in {}'.format(folder))

    volumes = []
    datasets = []

    file_names = file_names[slice_start:slice_end]

    for file_name in tqdm.tqdm(file_names, 'loading volume data'):
        # read the file
        dataset = dicom.read_file(folder + '/' + file_name)

        # Get parameters
        pixel_size = np.array(dataset.PixelSpacing)
        pixel_thickness = float(dataset.SliceThickness)
        rows = dataset.Rows
        cols = dataset.Columns

        # Get data array and convert to correct coordinates
        data_array = np.array(np.frombuffer(dataset.PixelData, 'H'),
                              dtype='float32')
        data_array = data_array.reshape([cols, rows], order='C')
        data_array = np.rot90(data_array, -1)

        # Convert from storage type to densities
        # TODO: Optimize these computations
        hu_values = (dataset.RescaleSlope * data_array +
                     dataset.RescaleIntercept)
        densities = (hu_values + 1000) / 1000

        # Store results
        volumes.append(densities)
        datasets.append(dataset)

    voxel_size = np.array(list(pixel_size) + [pixel_thickness])
    shape = np.array([rows, cols, len(volumes)])

    # Compute geometry parameters
    mid_pt = (np.array(dataset.ReconstructionTargetCenterPatient) -
              np.array(dataset.DataCollectionCenterPatient))
    reconstruction_size = (voxel_size * shape)
    min_pt = mid_pt - reconstruction_size / 2
    max_pt = mid_pt + reconstruction_size / 2

    # axis 1 has reversed convention
    min_pt[1], max_pt[1] = -max_pt[1], -min_pt[1]

    if len(datasets) > 1:
        slice_distance = np.abs(
            float(datasets[1].DataCollectionCenterPatient[2]) -
            float(datasets[0].DataCollectionCenterPatient[2]))
    else:
        # If we only have one slice, we must approximate the distance.
        slice_distance = pixel_thickness

    # The middle of the minimum/maximum slice can be computed from the
    # DICOM attribute "DataCollectionCenterPatient". Since ODL uses corner
    # points (e.g. edge of volume) we need to add half a voxel thickness to
    # both sides.
    min_pt[2] = -np.array(datasets[0].DataCollectionCenterPatient)[2]
    min_pt[2] -= 0.5 * slice_distance
    max_pt[2] = -np.array(datasets[-1].DataCollectionCenterPatient)[2]
    max_pt[2] += 0.5 * slice_distance

    partition = odl.uniform_partition(min_pt, max_pt, shape)

    volume = np.transpose(np.array(volumes), (1, 2, 0))

    return partition, volume


if __name__ == '__main__':
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import print_function, division, absolute_import
import numpy as np

__all__ = ('euler_matrix', 'axis_rotation', 'axis_rotation_matrix',
           'rotation_matrix_from_to', 'transform_system',
           'perpendicular_vector', 'is_inside_bounds')


def euler_matrix(phi, theta=None, psi=None):
    """Rotation matrix in 2 and 3 dimensions.

    Its rows represent the canonical unit vectors as seen from the
    rotated system while the columns are the rotated unit vectors as
    seen from the canonical system.

    Parameters
    ----------
    phi : float or `array-like`
        Either 2D counter-clockwise rotation angle (in radians) or first
        Euler angle.
    theta, psi : float or `array-like`, optional
        Second and third Euler angles in radians. If both are ``None``, a
        2D rotation matrix is computed. Otherwise a 3D rotation is computed,
        where the default ``None`` is equivalent to ``0.0``.
        The rotation is performed in "ZXZ" rotation order, see the
        Wikipedia article `Euler angles`_.

    Returns
    -------
    mat : `numpy.ndarray`
        Rotation matrix corresponding to the given angles. The
        returned array has shape ``(ndim, ndim)`` if all angles represent
        single parameters, with ``ndim == 2`` for ``phi`` only and
        ``ndim == 3`` for 2 or 3 Euler angles.
        If any of the angle parameters is an array, the shape of the
        returned array is ``broadcast(phi, theta, psi).shape + (ndim, ndim)``.

    References
    ----------
    .. _Euler angles:
        https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
    """
    if theta is None and psi is None:
        squeeze_out = (np.shape(phi) == ())
        ndim = 2
        phi = np.array(phi, dtype=float, copy=False, ndmin=1)
        theta = psi = 0.0
    else:
        # `None` broadcasts like a scalar
        squeeze_out = (np.broadcast(phi, theta, psi).shape == ())
        ndim = 3
        phi = np.array(phi, dtype=float, copy=False, ndmin=1)
        if theta is None:
            theta = 0.0
        if psi is None:
            psi = 0.0
        theta = np.array(theta, dtype=float, copy=False, ndmin=1)
        psi = np.array(psi, dtype=float, copy=False, ndmin=1)
        ndim = 3

    cph = np.cos(phi)
    sph = np.sin(phi)
    cth = np.cos(theta)
    sth = np.sin(theta)
    cps = np.cos(psi)
    sps = np.sin(psi)

    if ndim == 2:
        mat = np.array([[cph, -sph],
                        [sph, cph]])
    else:
        mat = np.array([
            [cph * cps - sph * cth * sps,
             -cph * sps - sph * cth * cps,
             sph * sth],
            [sph * cps + cph * cth * sps,
             -sph * sps + cph * cth * cps,
             -cph * sth],
            [sth * sps + 0 * cph,
             sth * cps + 0 * cph,
             cth + 0 * (cph + cps)]])  # Make sure all components broadcast

    if squeeze_out:
        return mat.squeeze()
    else:
        # Move the `(ndim, ndim)` axes to the end
        extra_dims = len(np.broadcast(phi, theta, psi).shape)
        newaxes = list(range(2, 2 + extra_dims)) + [0, 1]
        return np.transpose(mat, newaxes)


def axis_rotation(axis, angle, vectors, axis_shift=(0, 0, 0)):
    """Rotate a vector or an array of vectors around an axis in 3d.

    The rotation is computed by `Rodrigues' rotation formula`_.

    Parameters
    ----------
    axis : `array-like`, shape ``(3,)``
        Rotation axis, assumed to be a unit vector.
    angle : float
        Angle of the counter-clockwise rotation.
    vectors : `array-like`, shape ``(3,)`` or ``(N, 3)``
        The vector(s) to be rotated.
    axis_shift : `array_like`, shape ``(3,)``, optional
        Shift the rotation center by this vector. Note that only shifts
        perpendicular to ``axis`` matter.

    Returns
    -------
    rot_vec : `numpy.ndarray`
        The rotated vector(s).

    References
    ----------
    .. _Rodrigues' rotation formula:
        https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula

    Examples
    --------
    Rotating around the third coordinate axis by and angle of 90 degrees:

    >>> axis = (0, 0, 1)
    >>> rot1 = axis_rotation(axis, angle=np.pi / 2, vectors=(1, 0, 0))
    >>> np.allclose(rot1, (0, 1, 0))
    True
    >>> rot2 = axis_rotation(axis, angle=np.pi / 2, vectors=(0, 1, 0))
    >>> np.allclose(rot2, (-1, 0, 0))
    True

    The rotation can be performed with shifted rotation center. A shift
    along the axis does not matter:

    >>> rot3 = axis_rotation(axis, angle=np.pi / 2, vectors=(1, 0, 0),
    ...                      axis_shift=(0, 0, 2))
    >>> np.allclose(rot3, (0, 1, 0))
    True

    The distance between the rotation center and the vector to be rotated
    determines the radius of the rotation circle:

    >>> # Rotation center in the point to be rotated, should do nothing
    >>> rot4 = axis_rotation(axis, angle=np.pi / 2, vectors=(1, 0, 0),
    ...                      axis_shift=(1, 0, 0))
    >>> np.allclose(rot4, (1, 0, 0))
    True
    >>> # Distance 2, thus rotates to (0, 2, 0) in the shifted system,
    >>> # resulting in (-1, 2, 0) from shifting back after rotating
    >>> rot5 = axis_rotation(axis, angle=np.pi / 2, vectors=(1, 0, 0),
    ...                      axis_shift=(-1, 0, 0))
    >>> np.allclose(rot5, (-1, 2, 0))
    True

    Rotation of multiple vectors can be done in bulk:

    >>> vectors = [[1, 0, 0], [0, 1, 0]]
    >>> rot = axis_rotation(axis, angle=np.pi / 2, vectors=vectors)
    >>> np.allclose(rot[0], (0, 1, 0))
    True
    >>> np.allclose(rot[1], (-1, 0, 0))
    True
    """
    rot_matrix = axis_rotation_matrix(axis, angle)
    vectors = np.asarray(vectors, dtype=float)
    if vectors.shape == (3,):
        vectors = vectors[None, :]
    elif vectors.ndim == 2 and vectors.shape[1] == 3:
        pass
    else:
        raise ValueError('`vectors` must have shape (3,) or (N, 3), got array '
                         'with shape {}'.format(vectors.shape))

    # Get `axis_shift` part that is perpendicular to `axis`
    axis_shift = np.asarray(axis_shift, dtype=float)
    axis = np.asarray(axis, dtype=float)
    axis_shift = axis_shift - axis.dot(axis_shift) * axis

    # Shift vectors with the negative of the axis shift to move the rotation
    # center to the origin. Then rotate and shift back.
    centered_vecs = vectors - axis_shift[None, :]
    # Need to transpose the vectors to make the axis of length 3 come first
    rot_vecs = rot_matrix.dot(centered_vecs.T).T
    return axis_shift[None, :] + rot_vecs


def axis_rotation_matrix(axis, angle):
    """Matrix of the rotation around an axis in 3d.

    The matrix is computed according to `Rodriguez' rotation formula`_.

    Parameters
    ----------
    axis : `array-like`, shape ``(3,)``
        Rotation axis, assumed to be a unit vector.
    angle : float or `array-like`
        Angle(s) of counter-clockwise rotation.

    Returns
    -------
    mat : `numpy.ndarray`, shape ``(3, 3)``
        The axis rotation matrix.

    References
    ----------
    .. _Rodriguez' rotation formula:
        https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula
    """
    scalar_out = (np.shape(angle) == ())
    axis = np.asarray(axis)
    if axis.shape != (3,):
        raise ValueError('`axis` shape must be (3,), got {}'
                         ''.format(axis.shape))

    angle = np.array(angle, dtype=float, copy=False, ndmin=1)

    cross_mat = np.array([[0, -axis[2], axis[1]],
                          [axis[2], 0, -axis[0]],
                          [-axis[1], axis[0], 0]])
    dy_mat = np.outer(axis, axis)
    id_mat = np.eye(3)
    cos_ang = np.cos(angle)
    sin_ang = np.sin(angle)

    # Add extra dimensions for broadcasting
    extra_dims = cos_ang.ndim
    mat_slc = (None,) * extra_dims + (slice(None), slice(None))
    ang_slc = (slice(None),) * extra_dims + (None, None)
    # Matrices will have shape (1, ..., 1, ndim, ndim)
    cross_mat = cross_mat[mat_slc]
    dy_mat = dy_mat[mat_slc]
    id_mat = id_mat[mat_slc]
    # Angle arrays will have shape (..., 1, 1)
    cos_ang = cos_ang[ang_slc]
    sin_ang = sin_ang[ang_slc]

    axis_mat = cos_ang * id_mat + (1. - cos_ang) * dy_mat + sin_ang * cross_mat
    if scalar_out:
        return axis_mat.squeeze()
    else:
        return axis_mat


def rotation_matrix_from_to(from_vec, to_vec):
    r"""Return a matrix that rotates ``from_vec`` to ``to_vec`` in 2d or 3d.

    Since a rotation from one vector to another in 3 dimensions has
    (at least) one degree of freedom, this function makes deliberate but
    still arbitrary choices to fix these free parameters. See Notes for
    details. For the applied formula in 3d, see `this Wikipedia page
    about Rodrigues' rotation formula
    <https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula>`_.

    Parameters
    ----------
    from_vec, to_vec : `array-like`, shape ``(2,)`` or ``(3,)``
        Vectors between which the returned matrix rotates. They should not
        be very close to zero or collinear.

    Returns
    -------
    matrix : `numpy.ndarray`, shape ``(2, 2)`` or ``(3, 3)``
        A matrix rotating ``from_vec`` to ``to_vec``. Note that the
        matrix does *not* include scaling, i.e. it is not guaranteed
        that ``matrix.dot(from_vec) == to_vec``.

    Examples
    --------
    In two dimensions, rotation is simple:

    >>> from_vec, to_vec = [1, 0], [1, 1]
    >>> mat = rotation_matrix_from_to(from_vec, to_vec)
    >>> to_vec_normalized = np.divide(to_vec, np.linalg.norm(to_vec))
    >>> np.allclose(mat.dot([1, 0]), to_vec_normalized)
    True
    >>> from_vec, to_vec = [1, 0], [-1, 1]
    >>> mat = rotation_matrix_from_to(from_vec, to_vec)
    >>> to_vec_normalized = np.divide(to_vec, np.linalg.norm(to_vec))
    >>> np.allclose(mat.dot([1, 0]), to_vec_normalized)
    True

    Rotation in 3d by less than ``pi``:

    >>> from_vec, to_vec = [1, 0, 0], [-1, 1, 0]
    >>> mat = rotation_matrix_from_to(from_vec, to_vec)
    >>> to_vec_normalized = np.divide(to_vec, np.linalg.norm(to_vec))
    >>> np.allclose(mat.dot([1, 0, 0]), to_vec_normalized)
    True

    Rotation by more than ``pi``:

    >>> from_vec, to_vec = [1, 0, 0], [-1, -1, 0]
    >>> mat = rotation_matrix_from_to(from_vec, to_vec)
    >>> to_vec_normalized = np.divide(to_vec, np.linalg.norm(to_vec))
    >>> np.allclose(mat.dot([1, 0, 0]), to_vec_normalized)
    True

    Notes
    -----
    In 3d, the matrix corresponds to a rotation around the normal vector
    :math:`\hat n = \hat u \times \hat v`, where :math:`\hat u` and
    :math:`\hat v` are the normalized versions of :math:`u`, the
    vector from which to rotate, and :math:`v`, the vector to which
    should be rotated.

    The rotation angle is determined as
    :math:`\alpha = \pm \arccos(\langle \hat u, \hat v \rangle)`.
    Its sign corresponds to the sign of
    :math:`\langle \hat b, \hat v\rangle`, where
    :math:`\hat b = \hat n \times \hat u` is the binormal vector.

    In the case that :math:`\hat u` and :math:`\hat v` are collinear,
    a perpendicular vector is chosen as :math:`\hat n = (1, 0, 0)` if
    :math:`v_1 = v_2 = 0`, else :math:`\hat n = (-v_2, v_1, v_3)`.
    The angle in this case is :math:`\alpha = 0` if
    :math:`\langle \hat u, \hat v \rangle > 0`, otherwise
    :math:`\alpha = \pi`.
    """
    from_vec, from_vec_in = (np.array(from_vec, dtype=float, copy=True),
                             from_vec)
    to_vec, to_vec_in = np.array(to_vec, dtype=float, copy=True), to_vec

    if from_vec.shape not in ((2,), (3,)):
        raise ValueError('`from_vec.shape` must be (2,) or (3,), got {}'
                         ''.format(from_vec.shape))
    if to_vec.shape not in ((2,), (3,)):
        raise ValueError('`to_vec.shape` must be (2,) or (3,), got {}'
                         ''.format(to_vec.shape))
    if from_vec.shape != to_vec.shape:
        raise ValueError('`from_vec.shape` and `to_vec.shape` not equal: '
                         '{} != {}'
                         ''.format(from_vec.shape, to_vec.shape))

    ndim = len(from_vec)

    # Normalize vectors
    from_vec_norm = np.linalg.norm(from_vec)
    if from_vec_norm < 1e-10:
        raise ValueError('`from_vec` {} too close to zero'.format(from_vec_in))
    from_vec /= from_vec_norm
    to_vec_norm = np.linalg.norm(to_vec)
    if to_vec_norm < 1e-10:
        raise ValueError('`to_vec` {} too close to zero'.format(to_vec_in))
    to_vec /= to_vec_norm

    if ndim == 2:
        dot = np.dot(from_vec, to_vec)
        from_rot = (-from_vec[1], from_vec[0])
        if dot == 0:
            angle = np.pi / 2 if np.dot(from_rot, to_vec) > 0 else -np.pi / 2
        elif np.array_equal(to_vec, -from_vec):
            angle = np.pi
        else:
            angle = (np.sign(np.dot(from_rot, to_vec)) *
                     np.arccos(np.dot(from_vec, to_vec)))
        return np.array([[np.cos(angle), -np.sin(angle)],
                         [np.sin(angle), np.cos(angle)]])

    elif ndim == 3:
        # Determine normal
        normal = np.cross(from_vec, to_vec)
        normal_norm = np.linalg.norm(normal)

        if normal_norm < 1e-10:
            # Collinear vectors, use perpendicular vector and angle = 0 or pi
            normal = perpendicular_vector(from_vec)
            angle = 0 if np.dot(from_vec, to_vec) > 0 else np.pi
            return axis_rotation_matrix(normal, angle)
        else:
            # Usual case, determine binormal and sign of rotation angle
            normal /= normal_norm
            binormal = np.cross(normal, from_vec)
            angle = (np.sign(np.dot(binormal, to_vec)) *
                     np.arccos(np.dot(from_vec, to_vec)))
            return axis_rotation_matrix(normal, angle)

    else:
        raise RuntimeError('bad ndim')


def transform_system(principal_vec, principal_default, other_vecs,
                     matrix=None):
    """Transform vectors with either ``matrix`` or based on ``principal_vec``.

    The logic of this function is as follows:

    - If ``matrix`` is not ``None``, transform ``principal_vec`` and
      all vectors in ``other_vecs`` by ``matrix``, ignoring
      ``principal_default``.
    - If ``matrix`` is ``None``, compute the rotation matrix from
      ``principal_default`` to ``principal_vec``, not including the
      dilation. Apply that rotation to all vectors in ``other_vecs``.

    **Note:** All vectors must have the same shape and match the shape
    of ``matrix`` if given.

    Parameters
    ----------
    principal_vec : `array-like`, shape ``(ndim,)``
        Vector that defines the transformation if ``matrix`` is not
        provided.
    principal_default : `array-like`, shape ``(ndim,)``
        Default value for ``principal_vec``. The deviation from this
        determines the transformation.
        If ``matrix`` is given, this has no effect.
    other_vecs : sequence of ``None`` or `array-like`'s with shape ``(ndim,)``
        The other vectors that should be transformed. ``None`` entries
        are just appended as-is.
    matrix : `array-like`, shape ``(ndim, ndim)``, optional
        Explicit transformation matrix to be applied to the vectors.
        It is allowed to include a constant scaling but shouldn't have
        strongly varying directional scaling (bad condition).

    Returns
    -------
    transformed_vecs : tuple of `numpy.ndarray`, shape ``(ndim,)``
        The transformed vectors. The first entry is (the transformed)
        ``principal_vec``, followed by the transformed ``other_vecs``.
        Thus the length of the tuple is ``len(other_vecs) + 1``.
    """
    transformed_vecs = []
    principal_vec = np.asarray(principal_vec, dtype=float)
    ndim = principal_vec.shape[0]

    if matrix is None:
        # Separate into dilation and rotation. The dilation is only used
        # for comparison, not in the final matrix.
        principal_default = np.asarray(principal_default, dtype=float)

        pr_norm = np.linalg.norm(principal_vec)
        pr_default_norm = np.linalg.norm(principal_default)

        if pr_default_norm == 0.0 and pr_norm != 0.0:
            raise ValueError('no transformation from {} to {}'
                             ''.format(principal_default, principal_vec))
        elif pr_norm == 0.0 and pr_default_norm != 0.0:
            raise ValueError('transformation from {} to {} is singular'
                             ''.format(principal_default, principal_vec))
        elif pr_norm == 0.0 and pr_default_norm == 0.0:
            dilation = 1.0
        else:
            dilation = (np.linalg.norm(principal_vec) /
                        np.linalg.norm(principal_default))

        # Determine the rotation part
        if np.allclose(principal_vec, dilation * principal_default):
            # Dilation only
            matrix = np.eye(ndim)
        else:
            matrix = rotation_matrix_from_to(principal_default, principal_vec)

        # This one goes straight in
        transformed_vecs.append(principal_vec)

    else:
        matrix = np.asarray(matrix, dtype=float)
        if matrix.shape != (ndim, ndim):
            raise ValueError('matrix shape must be {}, got {}'
                             ''.format((ndim, ndim), matrix.shape))

        # Check matrix condition
        svals = np.linalg.svd(matrix, compute_uv=False)
        condition = np.inf if 0.0 in svals else svals[0] / svals[-1]
        if condition > 1e6:
            raise np.linalg.LinAlgError(
                'matrix is badly conditioned: condition number is {}'
                ''.format(condition))

        transformed_vecs.append(matrix.dot(principal_vec))

    for vec in other_vecs:
        if vec is None:
            transformed_vecs.append(None)
        else:
            transformed_vecs.append(matrix.dot(vec))

    return tuple(transformed_vecs)


def is_rotation_matrix(mat, show_diff=False):
    from scipy.linalg import det, norm

    dim = mat.shape[0]
    if dim != mat.shape[1]:
        return False

    determ = det(mat)
    right_handed = (np.abs(determ - 1.) < 1E-10)
    orthonorm_diff = mat * mat.T - np.eye(dim)
    diff_norm = norm(orthonorm_diff, 2)
    orthonormal = (diff_norm < 1E-10)
    if not right_handed or not orthonormal:
        if show_diff:
            print('matrix S:\n', mat)
            print('det(S): ', determ)
            print('S*S.T - eye:\n', orthonorm_diff)
            print('2-norm of difference: ', diff_norm)
        return False
    return True


def angles_from_matrix(rot_matrix):
    if rot_matrix.shape == (2, 2):
        theta = np.atan2(rot_matrix[1, 0], rot_matrix[0, 0])
        return theta,
    elif rot_matrix.shape == (3, 3):
        if rot_matrix[2, 2] == 1.:  # cannot use last row and column
            theta = 0.
            # upper-left block is 2d rotation for phi + psi, so one needs
            # to be fixed
            psi = 0.
            phi = np.atan2(rot_matrix[1, 0], rot_matrix[0, 0])
            if phi < 0:
                phi += 2 * np.pi  # in [0, 2pi)
        else:
            phi = np.atan2(rot_matrix[0, 2], -rot_matrix[1, 2])
            psi = np.atan2(rot_matrix[2, 0], rot_matrix[2, 1])
            theta = np.acos(rot_matrix[2, 2])

            if phi < 0. or psi < 0.:
                phi += np.pi
                psi += np.pi
                theta = -theta

        return phi, theta, psi

    else:
        raise ValueError('shape of `rot_matrix` must be (2, 2) or (3, 3), '
                         'got {}'.format(rot_matrix.shape))


def to_lab_sys(vec_in_local_coords, local_sys):
    vec_in_local_coords = np.array(vec_in_local_coords)
    trafo_matrix = np.matrix(local_sys).T
    return np.dot(trafo_matrix, vec_in_local_coords)


def to_local_sys(vec_in_lab_coords, local_sys):
    vec_in_lab_coords = np.array(vec_in_lab_coords)
    trafo_matrix = np.matrix(local_sys)
    return np.dot(trafo_matrix, vec_in_lab_coords)


def perpendicular_vector(vec):
    """Return a vector perpendicular to ``vec``.

    Parameters
    ----------
    vec : `array-like`
        Vector(s) of arbitrary length. The axis along the vector components
        must come last.

    Returns
    -------
    perp_vec : `numpy.ndarray`
        Array of same shape as ``vec`` such that ``dot(vec, perp_vec) == 0``
        (along the last axis if there are multiple vectors).

    Examples
    --------
    Works in 2d:

    >>> perpendicular_vector([0, 1])
    array([-1.,  0.])
    >>> np.allclose(perpendicular_vector([1, 0]), [0, 1])  # would print -0
    True

    And in 3d:

    >>> perpendicular_vector([0, 1, 0])
    array([-1.,  0.,  0.])
    >>> perpendicular_vector([0, 0, 1])
    array([ 1.,  0.,  0.])
    >>> np.allclose(perpendicular_vector([1, 0, 0]), [0, 1, 0])
    True

    The function is vectorized, i.e., it can be called with multiple
    vectors at once (additional axes being added to the left):

    >>> perpendicular_vector([[0, 1, 0],
    ...                       [0, 0, 1]])  # 2 vectors
    array([[-1.,  0.,  0.],
           [ 1.,  0.,  0.]])
    >>> vecs = np.zeros((2, 3, 3))
    >>> vecs[..., 1] = 1  # (2, 3) array of vectors (0, 1, 0)
    >>> perpendicular_vector(vecs)
    array([[[-1.,  0.,  0.],
            [-1.,  0.,  0.],
            [-1.,  0.,  0.]],
    <BLANKLINE>
           [[-1.,  0.,  0.],
            [-1.,  0.,  0.],
            [-1.,  0.,  0.]]])
    """
    squeeze_out = (np.ndim(vec) == 1)
    vec = np.array(vec, dtype=float, copy=False, ndmin=2)

    if np.any(np.all(vec == 0, axis=-1)):
        raise ValueError('zero vector')

    result = np.zeros(vec.shape)
    cond = np.any(vec[..., :2] != 0, axis=-1)
    result[cond, 0] = -vec[cond, 1]
    result[cond, 1] = vec[cond, 0]
    result[~cond, 0] = 1
    result /= np.linalg.norm(result, axis=-1, keepdims=True)

    if squeeze_out:
        result = result.squeeze()

    return result


def is_inside_bounds(value, params):
    """Return ``True`` if ``value`` is contained in ``params``.

    This method supports broadcasting in the sense that for
    ``params.ndim >= 2``, if more than one value is given, the inputs
    are broadcast against each other.

    Parameters
    ----------
    value : `array-like`
        Value(s) to be checked. For several inputs, the final bool
        tells whether all inputs pass the check or not.
    params : `IntervalProd`
        Set in which the value is / the values are supposed to lie.

    Returns
    -------
    is_inside_bounds : bool
        ``True`` is all values lie in ``params``, ``False`` otherwise.

    Examples
    --------
    Check a single point:

    >>> params = odl.IntervalProd([0, 0], [1, 2])
    >>> is_inside_bounds([0, 0], params)
    True
    >>> is_inside_bounds([0, -1], params)
    False

    Using broadcasting:

    >>> pts_ax0 = np.array([0, 0, 1, 0, 1])[:, None]
    >>> pts_ax1 = np.array([2, 0, 1])[None, :]
    >>> is_inside_bounds([pts_ax0, pts_ax1], params)
    True
    >>> pts_ax1 = np.array([-2, 1])[None, :]
    >>> is_inside_bounds([pts_ax0, pts_ax1], params)
    False
    """
    if value in params:
        # Single parameter
        return True
    else:
        if params.ndim == 1:
            return params.contains_all(np.ravel(value))
        else:
            # Flesh out and flatten to check bounds
            bcast_value = np.broadcast_arrays(*value)
            stacked_value = np.vstack(bcast_value)
            flat_value = stacked_value.reshape(params.ndim, -1)
            return params.contains_all(flat_value)


if __name__ == '__main__':
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the alternating dual updates solver."""

from __future__ import division
import odl
from odl.solvers.nonsmooth import adupdates
from odl.solvers.nonsmooth.alternating_dual_updates import adupdates_simple

from odl.util.testutils import all_almost_equal


# Places for the accepted error when comparing results
HIGH_ACCURACY = 8
LOW_ACCURACY = 4


def test_adupdates():
    """Test if the adupdates solver handles the following problem correctly:

    ( 1  1/2 1/3 1/4) (x_1)   (a)
    (1/2 1/3 1/4 1/5) (x_2)   (b)
    (1/3 1/4 1/5 1/6) (x_3) = (c)
    (1/4 1/5 1/6 1/7) (x_4)   (d).

    The matrix is the ill-conditined Hilbert matrix, the inverse of which
    can be given in closed form. If we set

    (a)   (25/12)                    (x_1) = (1)
    (b)   (77/60)                    (x_2) = (1)
    (c) = (19/20)         then       (x_3) = (1)
    (d)   (319/420),                 (x_4) = (1).

    We solve the problem

    min ||Ax - b||^2 + TV(x) s.t. x >= 0

    for the matrix A, the r.h.s. b as above and the total variation TV, which
    is given as the (non-cyclic) sum of the distances of consecutive entries
    of the solution. The solution of this problem is clearly x = (1, 1, 1, 1),
    since it satisfies the additional constraint and minimizes both terms of
    the objective function.
    """

    mat1 = [[1, 1 / 2, 1 / 3, 1 / 4],
            [1 / 2, 1 / 3, 1 / 4, 1 / 5]]
    mat2 = [[1 / 3, 1 / 4, 1 / 5, 1 / 6],
            [1 / 4, 1 / 5, 1 / 6, 1 / 7]]

    # Create the linear operators
    mat1op = odl.MatrixOperator(mat1)
    mat2op = odl.MatrixOperator(mat2)
    domain = mat1op.domain
    tv1 = odl.MatrixOperator([[1.0, -1.0, 0.0, 0.0]])
    tv2 = odl.MatrixOperator([[0.0, 0.0, 1.0, -1.0]])
    tv3 = odl.MatrixOperator([[0.0, 1.0, -1.0, 0.0]])
    nneg = odl.IdentityOperator(domain)
    ops = [mat1op, mat2op, odl.BroadcastOperator(tv1, tv2), tv3, nneg]

    # Create inner stepsizes for linear operators
    mat1s = 1 / mat1op(mat1op.adjoint(mat1op.range.one()))
    mat2s = 1 / mat2op(mat2op.adjoint(mat2op.range.one()))
    tv1s = [0.5, 0.5]
    tv2s = 0.5
    nnegs = nneg.range.element([1.0, 1.0, 1.0, 1.0])
    inner_stepsizes = [mat1s, mat2s, tv1s, tv2s, nnegs]

    expected_solution = domain.element([1, 1, 1, 1])
    # Create right-hand-sides of the equation
    rhs1 = mat1op(expected_solution)
    rhs2 = mat2op(expected_solution)

    # Create the functionals
    fid1 = odl.solvers.L2NormSquared(mat1op.range).translated(rhs1)
    fid2 = odl.solvers.L2NormSquared(mat2op.range).translated(rhs2)
    reg1 = odl.solvers.L1Norm(tv1.range)
    reg2 = odl.solvers.L1Norm(tv2.range)
    reg3 = odl.solvers.L1Norm(tv3.range)
    ind = odl.solvers.IndicatorNonnegativity(nneg.range)
    funcs = [fid1, fid2, odl.solvers.SeparableSum(reg1, reg2), reg3, ind]

    # Start from zero
    x = tv1.domain.zero()
    x_simple = tv1.domain.zero()

    stepsize = 1.0
    niter = 10

    adupdates(x, funcs, ops, stepsize, inner_stepsizes, niter)
    adupdates_simple(x_simple, funcs, ops, stepsize,
                     inner_stepsizes, niter)
    assert all_almost_equal(x, x_simple)
    assert domain.dist(x, expected_solution) < 3e-2


if __name__ == '__main__':
    odl.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Testing utilities for the ``tomo`` subpackage."""

from __future__ import absolute_import, division, print_function


__all__ = (
    'skip_if_no_astra',
    'skip_if_no_astra_cuda',
    'skip_if_no_skimage',
)

try:
    import pytest

except ImportError:

    def identity(*args, **kwargs):
        if args and callable(args[0]):
            return args[0]
        else:
            return identity

    skip_if_no_astra = skip_if_no_astra_cuda = skip_if_no_skimage = identity

else:
    skip_if_no_astra = pytest.mark.skipif(
        'not odl.tomo.ASTRA_AVAILABLE',
        reason='ASTRA not available',
    )
    skip_if_no_astra_cuda = pytest.mark.skipif(
        'not odl.tomo.ASTRA_CUDA_AVAILABLE',
        reason='ASTRA CUDA not available',
    )
    skip_if_no_skimage = pytest.mark.skipif(
        'not odl.tomo.SKIMAGE_AVAILABLE',
        reason='skimage not available',
    )



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import numpy as np

import odl
from odl import vector
from odl.space.npy_tensors import NumpyTensor
from odl.util.testutils import all_equal


def test_vector_numpy():

    # Rn
    inp = [[1.0, 2.0, 3.0],
           [4.0, 5.0, 6.0]]

    x = vector(inp)
    assert isinstance(x, NumpyTensor)
    assert x.dtype == np.dtype('float64')
    assert all_equal(x, inp)

    x = vector([1.0, 2.0, float('inf')])
    assert x.dtype == np.dtype('float64')
    assert isinstance(x, NumpyTensor)

    x = vector([1.0, 2.0, float('nan')])
    assert x.dtype == np.dtype('float64')
    assert isinstance(x, NumpyTensor)

    x = vector([1, 2, 3], dtype='float32')
    assert x.dtype == np.dtype('float32')
    assert isinstance(x, NumpyTensor)

    # Cn
    inp = [[1 + 1j, 2, 3 - 2j],
           [4 + 1j, 5, 6 - 1j]]

    x = vector(inp)
    assert isinstance(x, NumpyTensor)
    assert x.dtype == np.dtype('complex128')
    assert all_equal(x, inp)

    x = vector([1, 2, 3], dtype='complex64')
    assert isinstance(x, NumpyTensor)

    # Generic TensorSpace
    inp = [1, 2, 3]
    x = vector(inp)
    assert isinstance(x, NumpyTensor)
    assert x.dtype == np.dtype('int')
    assert all_equal(x, inp)

    inp = ['a', 'b', 'c']
    x = vector(inp)
    assert isinstance(x, NumpyTensor)
    assert np.issubdtype(x.dtype, np.str_)
    assert all_equal(x, inp)

    x = vector([1, 2, 'inf'])  # Becomes string type
    assert isinstance(x, NumpyTensor)
    assert np.issubdtype(x.dtype, np.str_)
    assert all_equal(x, ['1', '2', 'inf'])

    # Scalar or empty input
    x = vector(5.0)  # becomes 1d, size 1
    assert x.shape == (1,)

    x = vector([])  # becomes 1d, size 0
    assert x.shape == (0,)


if __name__ == '__main__':
    odl.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Base classes for implementations of tensor spaces."""

from __future__ import absolute_import, division, print_function

from numbers import Integral

import numpy as np

from odl.set.sets import ComplexNumbers, RealNumbers
from odl.set.space import LinearSpace, LinearSpaceElement
from odl.util import (
    array_str, dtype_str, indent, is_complex_floating_dtype, is_floating_dtype,
    is_numeric_dtype, is_real_dtype, is_real_floating_dtype, safe_int_conv,
    signature_string, writable_array)
from odl.util.ufuncs import TensorSpaceUfuncs
from odl.util.utility import TYPE_MAP_C2R, TYPE_MAP_R2C, nullcontext

__all__ = ('TensorSpace',)


class TensorSpace(LinearSpace):

    """Base class for sets of tensors of arbitrary data type.

    A tensor is, in the most general sense, a multi-dimensional array
    that allows operations per entry (keep the rank constant),
    reductions / contractions (reduce the rank) and broadcasting
    (raises the rank).
    For non-numeric data type like ``object``, the range of valid
    operations is rather limited since such a set of tensors does not
    define a vector space.
    Any numeric data type, on the other hand, is considered valid for
    a tensor space, although certain operations - like division with
    integer dtype - are not guaranteed to yield reasonable results.

    Under these restrictions, all basic vector space operations are
    supported by this class, along with reductions based on arithmetic
    or comparison, and element-wise mathematical functions ("ufuncs").

    See the `Wikipedia article on tensors`_ for further details.
    See also [Hac2012] "Part I Algebraic Tensors" for a rigorous
    treatment of tensors with a definition close to this one.

    Note also that this notion of tensors is the same as in popular
    Deep Learning frameworks.

    References
    ----------
    [Hac2012] Hackbusch, W. *Tensor Spaces and Numerical Tensor Calculus*.
    Springer, 2012.

    .. _Wikipedia article on tensors: https://en.wikipedia.org/wiki/Tensor
    """

    def __init__(self, shape, dtype):
        """Initialize a new instance.

        Parameters
        ----------
        shape : nonnegative int or sequence of nonnegative ints
            Number of entries of type ``dtype`` per axis in this space. A
            single integer results in a space with rank 1, i.e., 1 axis.
        dtype :
            Data type of elements in this space. Can be provided
            in any way the `numpy.dtype` constructor understands, e.g.
            as built-in type or as a string.
            For a data type with a ``dtype.shape``, these extra dimensions
            are added *to the left* of ``shape``.
        """
        # Handle shape and dtype, taking care also of dtypes with shape
        try:
            shape, shape_in = tuple(safe_int_conv(s) for s in shape), shape
        except TypeError:
            shape, shape_in = (safe_int_conv(shape),), shape
        if any(s < 0 for s in shape):
            raise ValueError('`shape` must have only nonnegative entries, got '
                             '{}'.format(shape_in))
        dtype = np.dtype(dtype)

        # We choose this order in contrast to Numpy, since we usually want
        # to represent discretizations of vector- or tensor-valued functions,
        # i.e., if dtype.shape == (3,) we expect f[0] to have shape `shape`.
        self.__shape = dtype.shape + shape
        self.__dtype = dtype.base

        if is_real_dtype(self.dtype):
            # real includes non-floating-point like integers
            field = RealNumbers()
            self.__real_dtype = self.dtype
            self.__real_space = self
            self.__complex_dtype = TYPE_MAP_R2C.get(self.dtype, None)
            self.__complex_space = None  # Set in first call of astype
        elif is_complex_floating_dtype(self.dtype):
            field = ComplexNumbers()
            self.__real_dtype = TYPE_MAP_C2R[self.dtype]
            self.__real_space = None  # Set in first call of astype
            self.__complex_dtype = self.dtype
            self.__complex_space = self
        else:
            field = None

        LinearSpace.__init__(self, field)

    @property
    def impl(self):
        """Name of the implementation back-end of this tensor set.

        This property should be overridden by subclasses.
        """
        raise NotImplementedError('abstract method')

    @property
    def shape(self):
        """Number of scalar elements per axis.

        .. note::
            If `dtype` has a shape, we add it to the **left** of the given
            ``shape`` in the class creation. This is in contrast to NumPy,
            which adds extra axes to the **right**. We do this since we
            usually want to represent discretizations of vector- or
            tensor-valued functions by this, i.e., if
            ``dtype.shape == (3,)`` we expect ``f[0]`` to have shape
            ``shape``.
        """
        return self.__shape

    @property
    def dtype(self):
        """Scalar data type of each entry in an element of this space."""
        return self.__dtype

    @property
    def is_real(self):
        """True if this is a space of real tensors."""
        return is_real_floating_dtype(self.dtype)

    @property
    def is_complex(self):
        """True if this is a space of complex tensors."""
        return is_complex_floating_dtype(self.dtype)

    @property
    def real_dtype(self):
        """The real dtype corresponding to this space's `dtype`.

        Raises
        ------
        NotImplementedError
            If `dtype` is not a numeric data type.
        """
        if not is_numeric_dtype(self.dtype):
            raise NotImplementedError(
                '`real_dtype` not defined for non-numeric `dtype`')
        return self.__real_dtype

    @property
    def complex_dtype(self):
        """The complex dtype corresponding to this space's `dtype`.

        Raises
        ------
        NotImplementedError
            If `dtype` is not a numeric data type.
        """
        if not is_numeric_dtype(self.dtype):
            raise NotImplementedError(
                '`complex_dtype` not defined for non-numeric `dtype`')
        return self.__complex_dtype

    @property
    def real_space(self):
        """The space corresponding to this space's `real_dtype`.

        Raises
        ------
        ValueError
            If `dtype` is not a numeric data type.
        """
        if not is_numeric_dtype(self.dtype):
            raise ValueError(
                '`real_space` not defined for non-numeric `dtype`')
        return self.astype(self.real_dtype)

    @property
    def complex_space(self):
        """The space corresponding to this space's `complex_dtype`.

        Raises
        ------
        ValueError
            If `dtype` is not a numeric data type.
        """
        if not is_numeric_dtype(self.dtype):
            raise ValueError(
                '`complex_space` not defined for non-numeric `dtype`')
        return self.astype(self.complex_dtype)

    def _astype(self, dtype):
        """Internal helper for `astype`.

        Subclasses with differing init parameters should overload this
        method.
        """
        kwargs = {}
        if is_floating_dtype(dtype):
            # Use weighting only for floating-point types, otherwise, e.g.,
            # `space.astype(bool)` would fail
            weighting = getattr(self, 'weighting', None)
            if weighting is not None:
                kwargs['weighting'] = weighting

        return type(self)(self.shape, dtype=dtype, **kwargs)

    def astype(self, dtype):
        """Return a copy of this space with new ``dtype``.

        Parameters
        ----------
        dtype :
            Scalar data type of the returned space. Can be provided
            in any way the `numpy.dtype` constructor understands, e.g.
            as built-in type or as a string. Data types with non-trivial
            shapes are not allowed.

        Returns
        -------
        newspace : `TensorSpace`
            Version of this space with given data type.
        """
        if dtype is None:
            # Need to filter this out since Numpy iterprets it as 'float'
            raise ValueError('`None` is not a valid data type')

        dtype = np.dtype(dtype)
        if dtype == self.dtype:
            return self

        if is_numeric_dtype(self.dtype):
            # Caching for real and complex versions (exact dtype mappings)
            if dtype == self.__real_dtype:
                if self.__real_space is None:
                    self.__real_space = self._astype(dtype)
                return self.__real_space
            elif dtype == self.__complex_dtype:
                if self.__complex_space is None:
                    self.__complex_space = self._astype(dtype)
                return self.__complex_space
            else:
                return self._astype(dtype)
        else:
            return self._astype(dtype)

    @property
    def default_order(self):
        """Default storage order for new elements in this space.

        This property should be overridden by subclasses.
        """
        raise NotImplementedError('abstract method')

    @property
    def size(self):
        """Total number of entries in an element of this space."""
        return (0 if self.shape == () else
                int(np.prod(self.shape, dtype='int64')))

    @property
    def ndim(self):
        """Number of axes (=dimensions) of this space, also called "rank"."""
        return len(self.shape)

    def __len__(self):
        """Number of tensor entries along the first axis."""
        return int(self.shape[0])

    @property
    def itemsize(self):
        """Size in bytes of one entry in an element of this space."""
        return int(self.dtype.itemsize)

    @property
    def nbytes(self):
        """Total number of bytes in memory used by an element of this space."""
        return self.size * self.itemsize

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contains : bool
            ``True`` if ``other`` has a ``space`` attribute that is equal
            to this space, ``False`` otherwise.

        Examples
        --------
        Elements created with the `TensorSpace.element` method are
        guaranteed to be contained in the same space:

        >>> spc = odl.tensor_space((2, 3), dtype='uint64')
        >>> spc.element() in spc
        True
        >>> x = spc.element([[0, 1, 2],
        ...                  [3, 4, 5]])
        >>> x in spc
        True

        Sizes, data types and other essential properties characterize
        spaces and decide about membership:

        >>> smaller_spc = odl.tensor_space((2, 2), dtype='uint64')
        >>> y = smaller_spc.element([[0, 1],
        ...                          [2, 3]])
        >>> y in spc
        False
        >>> x in smaller_spc
        False
        >>> other_dtype_spc = odl.tensor_space((2, 3), dtype='uint32')
        >>> z = other_dtype_spc.element([[0, 1, 2],
        ...                              [3, 4, 5]])
        >>> z in spc
        False
        >>> x in other_dtype_spc
        False

        On the other hand, spaces are not unique:

        >>> spc2 = odl.tensor_space((2, 3), dtype='uint64')
        >>> spc2 == spc
        True
        >>> x2 = spc2.element([[5, 4, 3],
        ...                    [2, 1, 0]])
        >>> x2 in spc
        True
        >>> x in spc2
        True

        Of course, random garbage is not in the space:

        >>> spc = odl.tensor_space((2, 3), dtype='uint64')
        >>> None in spc
        False
        >>> object in spc
        False
        >>> False in spc
        False
        """
        return getattr(other, 'space', None) == self

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            True if ``self`` and ``other`` have the same type, `shape`
            and `dtype`, otherwise ``False``.

        Examples
        --------
        Sizes, data types and other essential properties characterize
        spaces and decide about equality:

        >>> spc = odl.tensor_space(3, dtype='uint64')
        >>> spc == spc
        True
        >>> spc2 = odl.tensor_space(3, dtype='uint64')
        >>> spc2 == spc
        True
        >>> smaller_spc = odl.tensor_space(2, dtype='uint64')
        >>> spc == smaller_spc
        False
        >>> other_dtype_spc = odl.tensor_space(3, dtype='uint32')
        >>> spc == other_dtype_spc
        False
        >>> other_shape_spc = odl.tensor_space((3, 1), dtype='uint64')
        >>> spc == other_shape_spc
        False
        """
        if other is self:
            return True

        return (type(other) is type(self) and
                self.shape == other.shape and
                self.dtype == other.dtype)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), self.shape, self.dtype))

    def __repr__(self):
        """Return ``repr(self)``."""
        posargs = [self.shape, dtype_str(self.dtype)]
        return "{}({})".format(self.__class__.__name__,
                               signature_string(posargs, []))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)

    @property
    def examples(self):
        """Return example random vectors."""
        # Always return the same numbers
        rand_state = np.random.get_state()
        np.random.seed(1337)

        if is_numeric_dtype(self.dtype):
            yield ('Linearly spaced samples', self.element(
                np.linspace(0, 1, self.size).reshape(self.shape)))
            yield ('Normally distributed noise',
                   self.element(np.random.standard_normal(self.shape)))

        if self.is_real:
            yield ('Uniformly distributed noise',
                   self.element(np.random.uniform(size=self.shape)))
        elif self.is_complex:
            yield ('Uniformly distributed noise',
                   self.element(np.random.uniform(size=self.shape) +
                                np.random.uniform(size=self.shape) * 1j))
        else:
            # TODO: return something that always works, like zeros or ones?
            raise NotImplementedError('no examples available for non-numeric'
                                      'data type')

        np.random.set_state(rand_state)

    def zero(self):
        """Return a tensor of all zeros.

        This method should be overridden by subclasses.

        Returns
        -------
        zero : `Tensor`
            A tensor of all zeros.
        """
        raise NotImplementedError('abstract method')

    def one(self):
        """Return a tensor of all ones.

        This method should be overridden by subclasses.

        Returns
        -------
        one : `Tensor`
            A tensor of all one.
        """
        raise NotImplementedError('abstract method')

    def _multiply(self, x1, x2, out):
        """The entry-wise product of two tensors, assigned to ``out``.

        This method should be overridden by subclasses.
        """
        raise NotImplementedError('abstract method')

    def _divide(self, x1, x2, out):
        """The entry-wise quotient of two tensors, assigned to ``out``.

        This method should be overridden by subclasses.
        """
        raise NotImplementedError('abstract method')

    @staticmethod
    def default_dtype(field=None):
        """Return the default data type for a given field.

        This method should be overridden by subclasses.

        Parameters
        ----------
        field : `Field`, optional
            Set of numbers to be represented by a data type.

        Returns
        -------
        dtype :
            Numpy data type specifier.
        """
        raise NotImplementedError('abstract method')

    @staticmethod
    def available_dtypes():
        """Return the set of data types available in this implementation.

        This method should be overridden by subclasses.
        """
        raise NotImplementedError('abstract method')

    @property
    def element_type(self):
        """Type of elements in this space: `Tensor`."""
        return Tensor


class Tensor(LinearSpaceElement):

    """Abstract class for representation of `TensorSpace` elements."""

    def asarray(self, out=None):
        """Extract the data of this tensor as a Numpy array.

        This method should be overridden by subclasses.

        Parameters
        ----------
        out : `numpy.ndarray`, optional
            Array to write the result to.

        Returns
        -------
        asarray : `numpy.ndarray`
            Numpy array of the same data type and shape as the space.
            If ``out`` was given, the returned object is a reference
            to it.
        """
        raise NotImplementedError('abstract method')

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        This method should be overridden by subclasses.

        Parameters
        ----------
        indices : index expression
            Integer, slice or sequence of these, defining the positions
            of the data array which should be accessed.

        Returns
        -------
        values : `TensorSpace.dtype` or `Tensor`
            The value(s) at the given indices. Note that depending on
            the implementation, the returned object may be a (writable)
            view into the original array.
        """
        raise NotImplementedError('abstract method')

    def __setitem__(self, indices, values):
        """Implement ``self[indices] = values``.

        This method should be overridden by subclasses.

        Parameters
        ----------
        indices : index expression
            Integer, slice or sequence of these, defining the positions
            of the data array which should be written to.
        values : scalar, `array-like` or `Tensor`
            The value(s) that are to be assigned.

            If ``index`` is an integer, ``value`` must be a scalar.

            If ``index`` is a slice or a sequence of slices, ``value``
            must be broadcastable to the shape of the slice.
        """
        raise NotImplementedError('abstract method')

    @property
    def impl(self):
        """Name of the implementation back-end of this tensor."""
        return self.space.impl

    @property
    def shape(self):
        """Number of elements per axis."""
        return self.space.shape

    @property
    def dtype(self):
        """Data type of each entry."""
        return self.space.dtype

    @property
    def size(self):
        """Total number of entries."""
        return self.space.size

    @property
    def ndim(self):
        """Number of axes (=dimensions) of this tensor."""
        return self.space.ndim

    def __len__(self):
        """Return ``len(self)``.

        The length is equal to the number of entries along axis 0.
        """
        return len(self.space)

    @property
    def itemsize(self):
        """Size in bytes of one tensor entry."""
        return self.space.itemsize

    @property
    def nbytes(self):
        """Total number of bytes in memory occupied by this tensor."""
        return self.space.nbytes

    def astype(self, dtype):
        """Return a copy of this element with new ``dtype``.

        Parameters
        ----------
        dtype :
            Scalar data type of the returned space. Can be provided
            in any way the `numpy.dtype` constructor understands, e.g.
            as built-in type or as a string. Data types with non-trivial
            shapes are not allowed.

        Returns
        -------
        newelem : `Tensor`
            Version of this element with given data type.
        """
        raise NotImplementedError('abstract method')

    def __repr__(self):
        """Return ``repr(self)``."""
        maxsize_full_print = 2 * np.get_printoptions()['edgeitems']
        self_str = array_str(self, nprint=maxsize_full_print)
        if self.ndim == 1 and self.size <= maxsize_full_print:
            return '{!r}.element({})'.format(self.space, self_str)
        else:
            return '{!r}.element(\n{}\n)'.format(self.space, indent(self_str))

    def __str__(self):
        """Return ``str(self)``."""
        return array_str(self)

    def __bool__(self):
        """Return ``bool(self)``."""
        if self.size > 1:
            raise ValueError('The truth value of an array with more than one '
                             'element is ambiguous. '
                             'Use np.any(a) or np.all(a)')
        else:
            return bool(self.asarray())

    def __array__(self, dtype=None):
        """Return a Numpy array from this tensor.

        Parameters
        ----------
        dtype :
            Specifier for the data type of the output array.

        Returns
        -------
        array : `numpy.ndarray`
        """
        if dtype is None:
            return self.asarray()
        else:
            return self.asarray().astype(dtype, copy=False)

    def __array_wrap__(self, array):
        """Return a new tensor wrapping the ``array``.

        Parameters
        ----------
        array : `numpy.ndarray`
            Array to be wrapped.

        Returns
        -------
        wrapper : `Tensor`
            Tensor wrapping ``array``.
        """
        if array.ndim == 0:
            return self.space.field.element(array)
        else:
            return self.space.element(array)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Interface to Numpy's ufunc machinery.

        This method is called by Numpy version 1.13 and higher as a single
        point for the ufunc dispatch logic. An object implementing
        ``__array_ufunc__`` takes over control when a `numpy.ufunc` is
        called on it, allowing it to use custom implementations and
        output types.

        This includes handling of in-place arithmetic like
        ``npy_array += custom_obj``. In this case, the custom object's
        ``__array_ufunc__`` takes precedence over the baseline
        `numpy.ndarray` implementation. It will be called with
        ``npy_array`` as ``out`` argument, which ensures that the
        returned object is a Numpy array. For this to work properly,
        ``__array_ufunc__`` has to accept Numpy arrays as ``out`` arguments.

        See the `corresponding NEP`_ and the `interface documentation`_
        for further details. See also the `general documentation on
        Numpy ufuncs`_.

        .. note::
            This basic implementation casts inputs and
            outputs to Numpy arrays and evaluates ``ufunc`` on those.
            For `numpy.ndarray` based data storage, this incurs no
            significant overhead compared to direct usage of Numpy arrays.

            For other (in particular non-local) implementations, e.g.,
            GPU arrays or distributed memory, overhead is significant due
            to copies to CPU main memory. In those classes, the
            ``__array_ufunc__`` mechanism should be overridden in favor of
            a native implementations if possible.

        .. note::
            If no ``out`` parameter is provided, this implementation
            just returns the raw array and does not attempt to wrap the
            result in any kind of space.

        Parameters
        ----------
        ufunc : `numpy.ufunc`
            Ufunc that should be called on ``self``.
        method : str
            Method on ``ufunc`` that should be called on ``self``.
            Possible values:

            ``'__call__'``, ``'accumulate'``, ``'at'``, ``'outer'``,
            ``'reduce'``, ``'reduceat'``

        input1, ..., inputN:
            Positional arguments to ``ufunc.method``.
        kwargs:
            Keyword arguments to ``ufunc.method``.

        Returns
        -------
        ufunc_result : `Tensor`, `numpy.ndarray` or tuple
            Result of the ufunc evaluation. If no ``out`` keyword argument
            was given, the result is a `Tensor` or a tuple
            of such, depending on the number of outputs of ``ufunc``.
            If ``out`` was provided, the returned object or tuple entries
            refer(s) to ``out``.

        References
        ----------
        .. _corresponding NEP:
           https://docs.scipy.org/doc/numpy/neps/ufunc-overrides.html

        .. _interface documentation:
           https://docs.scipy.org/doc/numpy/reference/arrays.classes.html\
#numpy.class.__array_ufunc__

        .. _general documentation on Numpy ufuncs:
           https://docs.scipy.org/doc/numpy/reference/ufuncs.html

        .. _reduceat documentation:
           https://docs.scipy.org/doc/numpy/reference/generated/\
numpy.ufunc.reduceat.html
        """
        # --- Process `out` --- #

        # Unwrap out if provided. The output parameters are all wrapped
        # in one tuple, even if there is only one.
        out_tuple = kwargs.pop('out', ())

        # Check number of `out` args, depending on `method`
        if method == '__call__' and len(out_tuple) not in (0, ufunc.nout):
            raise ValueError(
                "ufunc {}: need 0 or {} `out` arguments for "
                "`method='__call__'`, got {}"
                ''.format(ufunc.__name__, ufunc.nout, len(out_tuple)))
        elif method != '__call__' and len(out_tuple) not in (0, 1):
            raise ValueError(
                'ufunc {}: need 0 or 1 `out` arguments for `method={!r}`, '
                'got {}'.format(ufunc.__name__, method, len(out_tuple)))

        # We allow our own tensors, the data container type and
        # `numpy.ndarray` objects as `out` (see docs for reason for the
        # latter)
        valid_types = (type(self), type(self.data), np.ndarray)
        if not all(isinstance(o, valid_types) or o is None
                   for o in out_tuple):
            return NotImplemented

        # Assign to `out` or `out1` and `out2`, respectively
        out = out1 = out2 = None
        if len(out_tuple) == 1:
            out = out_tuple[0]
        elif len(out_tuple) == 2:
            out1 = out_tuple[0]
            out2 = out_tuple[1]

        # --- Process `inputs` --- #

        # Convert inputs that are ODL tensors or their data containers to
        # Numpy arrays so that the native Numpy ufunc is called later
        inputs = tuple(
            np.asarray(inp) if isinstance(inp, (type(self), type(self.data)))
            else inp
            for inp in inputs)

        # --- Get some parameters for later --- #

        # Arguments for `writable_array` and/or space constructors
        out_dtype = kwargs.get('dtype', None)
        if out_dtype is None:
            array_kwargs = {}
        else:
            array_kwargs = {'dtype': out_dtype}

        # --- Evaluate ufunc --- #

        if method == '__call__':
            if ufunc.nout == 1:
                # Make context for output (trivial one returns `None`)
                if out is None:
                    out_ctx = nullcontext()
                else:
                    out_ctx = writable_array(out, **array_kwargs)

                # Evaluate ufunc
                with out_ctx as out_arr:
                    kwargs['out'] = out_arr
                    res = ufunc(*inputs, **kwargs)

                # Return result (may be a raw array or a space element)
                return res

            elif ufunc.nout == 2:
                # Make contexts for outputs (trivial ones return `None`)
                if out1 is not None:
                    out1_ctx = writable_array(out1, **array_kwargs)
                else:
                    out1_ctx = nullcontext()
                if out2 is not None:
                    out2_ctx = writable_array(out2, **array_kwargs)
                else:
                    out2_ctx = nullcontext()

                # Evaluate ufunc
                with out1_ctx as out1_arr, out2_ctx as out2_arr:
                    kwargs['out'] = (out1_arr, out2_arr)
                    res1, res2 = ufunc(*inputs, **kwargs)

                # Return results (may be raw arrays or space elements)
                return res1, res2

            else:
                raise NotImplementedError('nout = {} not supported'
                                          ''.format(ufunc.nout))

        else:  # method != '__call__'
            # Make context for output (trivial one returns `None`)
            if out is None:
                out_ctx = nullcontext()
            else:
                out_ctx = writable_array(out, **array_kwargs)

            # Evaluate ufunc method
            if method == 'at':
                with writable_array(inputs[0]) as inp_arr:
                    res = ufunc.at(inp_arr, *inputs[1:], **kwargs)
            else:
                with out_ctx as out_arr:
                    kwargs['out'] = out_arr
                    res = getattr(ufunc, method)(*inputs, **kwargs)

            # Return result (may be scalar, raw array or space element)
            return res

    # Old ufuncs interface, will be deprecated when Numpy 1.13 becomes minimum

    @property
    def ufuncs(self):
        """Access to Numpy style universal functions.

        These default ufuncs are always available, but may or may not be
        optimized for the specific space in use.

        .. note::
            This interface is will be deprecated when Numpy 1.13 becomes
            the minimum required version. Use Numpy ufuncs directly, e.g.,
            ``np.sqrt(x)`` instead of ``x.ufuncs.sqrt()``.
        """
        return TensorSpaceUfuncs(self)

    def show(self, title=None, method='', indices=None, force_show=False,
             fig=None, **kwargs):
        """Display the function graphically.

        Parameters
        ----------
        title : string, optional
            Set the title of the figure

        method : string, optional
            1d methods:

                ``'plot'`` : graph plot

                ``'scatter'`` : scattered 2d points (2nd axis <-> value)

            2d methods:

                ``'imshow'`` : image plot with coloring according to
                value, including a colorbar.

                ``'scatter'`` : cloud of scattered 3d points
                (3rd axis <-> value)

        indices : index expression, optional
            Display a slice of the array instead of the full array. The
            index expression is most easily created with the `numpy.s_`
            constructor, i.e. supply ``np.s_[:, 1, :]`` to display the
            first slice along the second axis.
            For data with 3 or more dimensions, the 2d slice in the first
            two axes at the "middle" along the remaining axes is shown
            (semantically ``[:, :, shape[2:] // 2]``).
            This option is mutually exclusive to ``coords``.

        force_show : bool, optional
            Whether the plot should be forced to be shown now or deferred until
            later. Note that some backends always displays the plot, regardless
            of this value.

        fig : `matplotlib.figure.Figure`, optional
            The figure to show in. Expected to be of same "style", as
            the figure given by this function. The most common use case
            is that ``fig`` is the return value of an earlier call to
            this function.

        kwargs : {'figsize', 'saveto', 'clim', ...}, optional
            Extra keyword arguments passed on to the display method.
            See the Matplotlib functions for documentation of extra
            options.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            The resulting figure. It is also shown to the user.

        See Also
        --------
        odl.util.graphics.show_discrete_data : Underlying implementation
        """
        from odl.discr import uniform_grid
        from odl.util.graphics import show_discrete_data

        # Default to showing x-y slice "in the middle"
        if indices is None and self.ndim >= 3:
            indices = tuple(
                [slice(None)] * 2 + [n // 2 for n in self.space.shape[2:]]
            )

        if isinstance(indices, (Integral, slice)):
            indices = (indices,)
        elif indices is None or indices == Ellipsis:
            indices = (slice(None),) * self.ndim
        else:
            indices = tuple(indices)

        # Replace None by slice(None)
        indices = tuple(slice(None) if idx is None else idx for idx in indices)

        if Ellipsis in indices:
            # Replace Ellipsis with the correct number of [:] expressions
            pos = indices.index(Ellipsis)
            indices = (indices[:pos] +
                       (np.s_[:], ) * (self.ndim - len(indices) + 1) +
                       indices[pos + 1:])

        if len(indices) < self.ndim:
            raise ValueError('too few axes ({} < {})'.format(len(indices),
                                                             self.ndim))
        if len(indices) > self.ndim:
            raise ValueError('too many axes ({} > {})'.format(len(indices),
                                                              self.ndim))

        # Squeeze grid and values according to the index expression
        full_grid = uniform_grid([0] * self.ndim, np.array(self.shape) - 1,
                                 self.shape)
        grid = full_grid[indices].squeeze()
        values = self.asarray()[indices].squeeze()

        return show_discrete_data(values, grid, title=title, method=method,
                                  force_show=force_show, fig=fig, **kwargs)


if __name__ == '__main__':
    from odl.util.testutils import run_doctests
    run_doctests()
